#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import http
import logging
import tornado.web
import tornado.gen
import tornado.escape
from .supports import singleton

LOG = logging.getLogger('tornadohandler')

@singleton
class AsyncRoutes():
    """
    """
    def __init__(self):
        self.routes = []

routes = AsyncRoutes()

class GeneralTornadoHandler(tornado.web.RequestHandler):
    """
    """

    def initialize(self, callback, methods, ac=[]):
        self.callbacks = {}
        self._ac = []
        if isinstance(methods, str):
            methods = [methods]
        for method in methods:
            method = method.upper()
            self.callbacks[method] = callback
        self.set_access_control(ac)

    def set_access_control(self, ac):
        self._ac = []
        if ac:
            if isinstance(ac, list):
                for f in ac:
                    if callable(f):
                        self._ac.append(['', f])
            elif isinstance(ac, dict):
                for k, f in ac:
                    if callable(f):
                        self._ac.append([k, f])
            elif callable(ac):
                self._ac.append(['', ac])

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        yield self._do_callback(self.callbacks.get('GET'), *args, **kwargs)

    @tornado.gen.coroutine
    def post(self, *args, **kwargs):
        yield self._do_callback(self.callbacks.get('POST'), *args, **kwargs)
    
    @tornado.gen.coroutine
    def _do_callback(self, cb, *args, **kwargs):
        # LOG.debug(' - %s %s', self.request.host_name, self.request.uri)
        if callable(cb):
            is_reject = False
            reject_resson = ''
            for o in self._ac:
                f = o[1]
                if tornado.gen.is_coroutine_function(f):
                    ac_result, msg = yield f(self.request)
                else:
                    ac_result, msg = f(self.request)
                if not ac_result:
                    reject_resson = str(msg)
                    LOG.warning('verify %s access control %s failed with error:%s', self.request.path, str(o[0]), reject_resson)
                    is_reject = True
                    break
            if is_reject:
                self.write(reject_resson)
                self.set_status(http.HTTPStatus.FORBIDDEN)
                self.finish()
                return

            argsx = [self, self.request]
            for arg in args:
                argsx.append(arg)

            # fc = cb.__code__
            # if fc.co_name == 'wrapper':
            if tornado.gen.is_coroutine_function(cb):
                response = yield cb(*argsx, **kwargs)
            else:
                response = cb(*argsx, **kwargs)

            if isinstance(response, tuple):
                status_code = response[1]
                response = response[0]
                if status_code is not None:
                    self.set_status(status_code)

            if isinstance(response, str):
                self.write(response)
                self.finish()
            else:
                LOG.error("====== unknown response type:%s", str(response))
                if not self._finished:
                    self.finish()
        else:
            print('not callable cb:', cb)
            self.set_status(http.HTTPStatus.METHOD_NOT_ALLOWED)
            self.finish()

class PageNotFoundHandler(tornado.web.RequestHandler):
    def get(self):
        raise tornado.web.HTTPError(404)

def async_route(rule, **options):
    def decorator(f):
        # endpoint = options.pop('endpoint', None)
        # if not endpoint:
        #     endpoint = f.__name__
        ac = options.pop('ac', [])

        routes.routes.append((rule, GeneralTornadoHandler, dict(callback=f, methods=options.pop('methods', ['GET']), ac=ac)))
        return f
    return decorator

def args_as_dict(request):
    if hasattr(request, '_args_as_dict'):
        return getattr(request, '_args_as_dict')
    args = {}
    for k, v in request.arguments.items():
        if len(v) == 1:
            v0 = v[0]
            if isinstance(v0, bytes):
                v0 = v0.decode()
            args[k] = v0
        else:
            args[k] = v
    setattr(request, '_args_as_dict', args)
    return args

def request_body_as_json(request):
    if hasattr(request, '_body_as_json'):
        return getattr(request, '_body_as_json')
    params = {}
    try:
        params = tornado.escape.json_decode(request.body)
        if params is None:
            params = {}
    except Exception as e:
        LOG.error('parse request body:%s failed with error:%s', request.body.decode(), str(e))
    
    setattr(request, '_body_as_json', params)
    return params

def register_static_route(static_url_path=r'static', static_folder='static'):
    dst_static_folder = os.path.abspath(static_folder)
    routes.routes.append((static_url_path+r'/(.*)', tornado.web.StaticFileHandler, dict(path=dst_static_folder)))

def register_failover_route():
    routes.routes.append((r'.*', PageNotFoundHandler))

def check_post_params(params, must_list, must_type=None, is_or=False):
    if not isinstance(must_list, list):
        return True
    if is_or:
        for field in must_list:
            if field in params:
                return True
        return "lack %s" % ','.join(must_list)
    if isinstance(params, list):
        for p in params:
            p_result = check_post_params(p, must_list)
            if isinstance(p_result, str):
                return p_result
    if isinstance(params, dict):
        for attr in must_list:
            if attr not in params:
                return 'lack %s' % attr
    if must_type and isinstance(must_type, dict):
        for k, v in must_type.items():
            if not isinstance(params.get(k), v):
                return "%s type error, need %s" % (k, str(v))
    return True



from tornado import gen

from zeep.wsdl import bindings

__all__ = ['AsyncSoap11Binding', 'AsyncSoap12Binding']

class AsyncSoapBinding(object):

    @gen.coroutine
    def send(self, client, options, operation, args, kwargs):
        envelope, http_headers = self._create(
            operation, args, kwargs,
            client=client,
            options=options)

        response = yield client.transport.post_xml(
            options['address'], envelope, http_headers)

        operation_obj = self.get(operation)
        raise gen.Return(self.process_reply(client, operation_obj, response))

class AsyncSoap11Binding(AsyncSoapBinding, bindings.Soap11Binding):
    pass

class AsyncSoap12Binding(AsyncSoapBinding, bindings.Soap12Binding):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import logging.handlers
import os
import time
from lycium.kafka.debugSet import LEVEL


class Logger(object):

    def __init__(self):
        self.logger = logging.getLogger("")
        # 设置输出的等级
        LEVELS = {'NOSET': logging.NOTSET,
                  'DEBUG': logging.DEBUG,
                  'INFO': logging.INFO,
                  'WARNING': logging.WARNING,
                  'ERROR': logging.ERROR,
                  'CRITICAL': logging.CRITICAL}
        self.logger.setLevel(LEVELS.get(LEVEL, 'INFO'))
        # 创建文件目录
        logs_dir = "logs"
        if os.path.exists(logs_dir) and os.path.isdir(logs_dir):
            pass
        else:
            os.mkdir(logs_dir)
        # 修改log保存位置
        timestamp = time.strftime("%Y-%m-%d", time.localtime())
        logfilename = '%s.txt' % timestamp
        logfilepath = os.path.join(logs_dir, logfilename)
        rotating_file_handler = logging.handlers.RotatingFileHandler(filename=logfilepath,
                                                                   maxBytes=1024 * 1024 * 50,
                                                                   backupCount=5)
        # 设置输出格式
        formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s', '%Y-%m-%d %H:%M:%S')
        rotating_file_handler.setFormatter(formatter)
        # 控制台句柄
        console = logging.StreamHandler()
        console.setLevel(LEVELS.get(LEVEL, 'INFO'))
        console.setFormatter(formatter)
        # 添加内容到日志句柄中
        self.logger.addHandler(rotating_file_handler)
        self.logger.addHandler(console)
        self.logger.setLevel(LEVELS.get(LEVEL, 'INFO'))

    def info(self, message):
        self.logger.info(message)

    def debug(self, message):
        self.logger.debug(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def exception(self, message):
        self.logger.exception(message)


logger = Logger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import urllib
import logging
from .http_utils import url_query_json

LOG = logging.getLogger("utilities.restful_api")

def restful_get(uri, qry):
    params = urllib.parse.urlencode({'filter':json.dumps(qry)})
    ret, result = url_query_json(uri+'?'+params)
    if ret:
        LOG.info("query uri:%s succeed. result:%s", uri, str(result))
        return result
    else:
        LOG.error("query uri:%s failed with error:%s", uri, result)
        return ret

def restful_post(uri, data):
    ret, result = url_query_json(uri, data)
    if ret:
        LOG.info("post uri:%s succeed. result:%s", uri, str(result))
        return result
    else:
        LOG.error("post uri:%s failed with error:%s", uri, result)
        return ret
        
def restful_patch(uri, data, params={}):
    if params:
        params = urllib.parse.urlencode(params)
        uri = uri + '?' + params
    ret, result = url_query_json(uri, data, 'PATCH')
    if ret:
        LOG.info("query uri:%s with method:patch succeed. result:%s", uri, str(result))
        return result
    else:
        LOG.error("query uri:%s with method:patch failed with error:%s", uri, result)
        return ret
        


#!/usr/bin/env python3.6
#! -*- coding: utf-8 -*-

import rsa
from . import base64_encode

def rsa_encode(data, pubkey, format='PEM'):
    if isinstance(pubkey, str):
        pubkey = rsa.PublicKey.load_pkcs1(pubkey.encode(), format=format)
    elif isinstance(pubkey, bytes):
        pubkey = rsa.PublicKey.load_pkcs1(pubkey, format=format)
    enc_text = rsa.encrypt(data.encode(), pubkey)
    return base64_encode(enc_text)

def rsa_encode_ne(data, pubkeys):
    pubkey = rsa.PublicKey(pubkeys.get('n'), pubkeys.get('e'))
    enc_text = rsa.encrypt(data.encode(), pubkey)
    return base64_encode(enc_text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import time
from lycium.exceptionreporter import ExceptionReporter

class TestExceptionReporter(unittest.TestCase):
    """
    """
    def testReport(self):
        ExceptionReporter().configure('test', {'connector': 'es', 'host': '127.0.0.1', 'port': 9200})
        ExceptionReporter().report(key='test', typ='unittest', 
            endpoint='debug+python://localhost/exceptionreporter',
            method='debug',
            inputs='{}',
            outputs='{"code":0, "message": "Debugging"}',
            content='foo',
            level='DEBUG'
        )

if __name__ == "__main__":
    unittest.main()
    time.sleep(10)




"""
Adds async tornado.gen support to Zeep.

"""
import logging
import urllib

from requests import Response, Session
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from tornado import gen, httpclient

from . import bindings
from zeep.transports import Transport
from zeep.utils import get_version
from zeep.wsdl.utils import etree_to_string

__all__ = ['TornadoAsyncTransport']

class TornadoAsyncTransport(Transport):
    """Asynchronous Transport class using tornado gen."""
    binding_classes = [
        bindings.AsyncSoap11Binding,
        bindings.AsyncSoap12Binding]

    def __init__(self, cache=None, timeout=300, operation_timeout=None,
                 session=None):
        self.cache = cache
        self.load_timeout = timeout
        self.operation_timeout = operation_timeout
        self.logger = logging.getLogger(__name__)

        self.session = session or Session()
        self.session.headers['User-Agent'] = (
            'Zeep/%s (www.python-zeep.org)' % (get_version()))

    def _load_remote_data(self, url):
        client = httpclient.HTTPClient()
        kwargs = {
            'method': 'GET',
            'connect_timeout': self.load_timeout,
            'request_timeout': self.load_timeout
        }
        http_req = httpclient.HTTPRequest(url, **kwargs)
        response = client.fetch(http_req)
        return response.body

    @gen.coroutine
    def post(self, address, message, headers):
        response = yield self.fetch(address, 'POST', headers, message)

        raise gen.Return(response)

    @gen.coroutine
    def post_xml(self, address, envelope, headers):
        message = etree_to_string(envelope)

        response = yield self.post(address, message, headers)

        raise gen.Return(response)

    @gen.coroutine
    def get(self, address, params, headers):
        if params:
            if isinstance(params):
                address += '?' + urllib.parse.quote(params)
            else:
                address += '?' + urllib.parse.urlencode(params)
        response = yield self.fetch(address, 'GET', headers)

        raise gen.Return(response)

    @gen.coroutine
    def fetch(self, address, method, headers, message=None):
        async_client = httpclient.AsyncHTTPClient()

        # extracting auth
        auth_username = None
        auth_password = None
        auth_mode = None

        if self.session.auth:
            if type(self.session.auth) is tuple:
                auth_username = self.session.auth[0]
                auth_password = self.session.auth[1]
                auth_mode = 'basic'
            elif type(self.session.auth) is HTTPBasicAuth:
                auth_username = self.session.username
                auth_password = self.session.password
                auth_mode = 'basic'
            elif type(self.session.auth) is HTTPDigestAuth:
                auth_username = self.session.username
                auth_password = self.session.password
                auth_mode = 'digest'
            else:
                raise Exception('Not supported authentication.')

        # extracting client cert
        client_cert = None
        client_key = None
        ca_certs = None

        if self.session.cert:
            if type(self.session.cert) is str:
                ca_certs = self.session.cert
            elif type(self.session.cert) is tuple:
                client_cert = self.session.cert[0]
                client_key = self.session.cert[1]

        session_headers = dict(self.session.headers.items())

        kwargs = {
            'method': method,
            'connect_timeout': self.operation_timeout,
            'request_timeout': self.operation_timeout,
            'headers': dict(headers, **session_headers),
            'auth_username': auth_username,
            'auth_password': auth_password,
            'auth_mode': auth_mode,
            'validate_cert': bool(self.session.verify),
            'ca_certs': ca_certs,
            'client_key': client_key,
            'client_cert': client_cert
        }

        if message:
            kwargs['body'] = message

        http_req = httpclient.HTTPRequest(address, **kwargs)
        response = yield async_client.fetch(http_req, raise_error=False)

        raise gen.Return(self.new_response(response))

    @staticmethod
    def new_response(response):
        """Convert an tornado.HTTPResponse object to a requests.Response object"""
        new = Response()
        new._content = response.body
        new.status_code = response.code
        new.headers = dict(response.headers.get_all())
        return new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import time
import logging
import sys
import tornado.gen
from tornado.ioloop import IOLoop
from apscheduler.schedulers.tornado import TornadoScheduler

from lycium.amqplib import RabbitMQFactory

def test_amqplib():
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    virtual_host = '/'
    example_exchange = 'ex.fanout.example'
    example_queue = 'example.text'

    @tornado.gen.coroutine
    def test_callback(unused_channel, basic_deliver, properties, body):
        time.sleep(0.5)
        print('== on message consumer_tag:%s delivery_tag:%s' % (basic_deliver.consumer_tag, basic_deliver.delivery_tag), body)
        return 'Response: ' + str(body)

    def test_publishing(example, idx_obj):
        example.publish(virtual_host, example_exchange, example_queue, 'test-'+str(idx_obj['index']))
        idx_obj['index'] += 1
        example.publish(virtual_host, example_exchange, example_queue, 'test-'+str(idx_obj['index']))
        idx_obj['index'] += 1

    example = RabbitMQFactory()
    example.initialize({
        'host':'localhost',
        'port':5672, 
        'username':'guest', 
        'password':'guest',
        'virtual_host': virtual_host
    })
    example.consume(virtual_host, example_exchange, 'fanout', example_queue, example_queue, False, test_callback)
    example.publish(virtual_host, example_exchange, example_queue, 'test-rpc')
    
    # scheduler example
    scheduler = TornadoScheduler()
    idx_obj = { 'index': 0 }
    scheduler.add_job(test_publishing, 'interval', seconds=5, args=(example, idx_obj))
    scheduler.start()

    # start rabbit mq connection
    example.run()

    # start the ioloop
    IOLoop.instance().start()

if __name__ == '__main__':
    test_amqplib()



#!/usr/bin/env python3.6
#! -*- coding: utf-8 -*-
import json


def readJson(fileName):
    inJson = open(fileName, 'r')
    data = json.load(inJson)
    inJson.close()
    return data


def writeJson(fileName, data):
    outJson = open(fileName, 'w')
    outJson.write(json.dumps(data))



#!/usr/bin/env python
# encoding: utf-8
from lycium.kafka.publishesMessage import PublishesMessage
# from publishesResponse import PublishesResponse
from lycium.kafka.protocol.kafkapacket_pb2 import KafkaPacket
from lycium.kafka.logger import logger
import json
import asyncio


class WorkerDelegate(object):
    """ 
    用于执行 回调函数以及对应的参数
    """

    def __init__(self, callback: callable, message:KafkaPacket , private_topic: str):
        self.callback = callback
        self.message = message
        self.private_topic = private_topic

    async def executor(self):
        """ 
        用于执行回调函数，返回PublishesMessage
        """
        headers = self.message.headers
        response_headers = dict()
        for item in headers:
            response_headers[item.name] = item.value
        publish_message = PublishesMessage(payload="",
                                           reply_to=self.private_topic,
                                           send_to=self.message.replyTo,
                                           correlation_id=self.message.correlationId,
                                           is_reply=True,
                                           content_type=self.message.contentType,
                                           content_encoding=self.message.contentEncoding,
                                           group_id=self.message.groupId,
                                           message_id=self.message.messageId,
                                           msg_type=self.message.type,
                                           user_id=self.message.userId,
                                           app_id=self.message.appId,
                                           headers=response_headers,
                                           callback=None
                                           )
        if callable(self.callback):
            result = ""
            try:

                if asyncio.iscoroutinefunction(self.callback):
                    result = await self.callback(self.message)
                else:
                    result = self.callback(self.message)
                if isinstance(result, (list, dict)):
                    try:
                        result = json.dumps(result)
                    except Exception as ee:
                        logger.exception(ee)
                        logger.error('json serialize %s failed with error:%s', str(result), str(ee))
                        result = ""
                        publish_message.status_code = 500
                        publish_message.error_message = str(ee)
                elif isinstance(result, (bytes, str)):
                    pass
                else:
                    result = ''
            except Exception as e:
                logger.exception(e)
                publish_message.status_code = 500
                publish_message.error_message = str(e)
            publish_message.payload = result
        return publish_message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time

def calc_timestamp_anchor_in_hour(curTs, interval):
    tslicecount = int(3600 / interval)
    if tslicecount <= 0:
        tslicecount = 1
    st = time.localtime(curTs)
    tmin = st.tm_min
    tsec = st.tm_sec
    if tslicecount > 60:
        step = int(60 / (tslicecount / 60))
        tsec = st.tm_sec - (st.tm_sec % step)
    else:
        step = int(60 / tslicecount)
        tsec = 0
        tmin = st.tm_min - (st.tm_min % step)
    cvtTs = time.mktime((st.tm_year, st.tm_mon, st.tm_mday, st.tm_hour, tmin, tsec, st.tm_wday, st.tm_yday, st.tm_isdst))

    return cvtTs



#!/usr/bin/env python
# coding: utf-8

class Helper(object):
    """
    辅助基类模块
    """

    def __init__(self):
        self._config = dict()

        self._config["session.timeout.ms"] = 6000
        self._config["heartbeat.interval.ms"] = 2000
        self._partition = 0
        # 发送消息的状态，True 是成功发送，False 是发送失败
        self._send_error = None
        self._send_msg = None

    def config_servers(self, servers: list):
        """
        配置连接的服务器,如['localhost:9092']
        """
        self._config["bootstrap.servers"] = ",".join(servers)
        return self

    def config_partition(self, partition: int):
        """
        配置分区，如partition 为0
        """
        self._partition = partition
        return self

    def config_kerberos_service_name(self, name: str):
        """
        配置 kerberos.service.name
        使用kerberos 认证需要配置
        """
        self._config['kerberos.service.name'] = name
        return self

    def config_kerberos_keytab(self, path: str):
        """
        配置kerberos.keytab
        使用kerberos 认证需要配置
        """
        self._config['kerberos.keytab'] = path
        return self

    def config_kerberos_principal(self, value: str):
        """
        配置kerberos.principal
        使用kerberos 认证需要配置
        """
        self._config['kerberos.principal'] = value
        return self

    def config_security_protocol(self, protocol='sasl_plaintext'):
        """
        配置security.protocol
        使用plain 和kerberos 认证需要配置,使用默认即可
        """
        self._config['security.protocol'] = protocol
        return self

    def config_sasl_mechanisms(self, mechanism='PLAIN'):
        """
        配置 sasl.mechanisms
        使用plain 认证需要配置
        """
        self._config['sasl.mechanisms'] = mechanism
        return self

    def config_sasl_username(self, name):
        """
        配置 sasl.username
        使用plain 认证需要配置
        """
        self._config['sasl.username'] = name
        return self

    def config_sasl_password(self, password):
        """
        配置 sasl.password
        使用plain 认证需要配置
        """
        self._config['sasl.password'] = password
        return self

    def config_reconnect_interval(self, interval:int=500):
        """
        配置断线重连的时间间隔，单位是毫秒，默认是500
        """
        self._config['reconnect.backoff.ms'] = interval


#!/bin/bash
# encoding: utf-8
""" 

"""

from lycium.kafka.kafkaWorker import KafkaWorker
import asyncio
import datetime
import json

hosts = ["localhost:9092", "localhost:9093", "localhost:9094"]


async def test_query(worker, message):
    print("start query {0}".format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
    resp = await worker.send(topic="test",
                             message=json.dumps(message).encode("utf-8"),
                             content_type='test',
                             content_encoding='GBK',
                             group_id='0',
                             message_id='1',
                             msg_type='msg_type',
                             user_id='user11',
                             app_id='10',
                             headers={'name1': '123', 'name2': '321'},
                             need_reply=True)
    print("end query {0}".format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
    print("client receive value:{0}".format(resp.body))


def run():
    kafak_worker = KafkaWorker(hosts, private_topic="client23")
    event_loop = kafak_worker.event_loop
    event_loop.call_later(1, event_loop.create_task, test_query(kafak_worker, {"action": "say"}))
    event_loop.call_later(2, event_loop.create_task, test_query(kafak_worker, {"action": "eat"}))
    event_loop.run_forever()


if __name__ == "__main__":
    run()



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

import re
import datetime, time
import logging
import random
import hashlib
import base64
import urllib

LOG = logging.getLogger('utilities')

_CHARACTORS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()_+|'

def singleton(clsname):
    instances = {}
    def getinstance(*args,**kwargs):
        if clsname not in instances:
            instances[clsname] = clsname(*args,**kwargs)
        return instances[clsname]
    return getinstance

class Constant(object):
    class ConstError(TypeError) : pass
    
    def __setattr__(self, key, value):
        if hasattr(self, key):
            raise self.ConstError("Can't rebind const (%s)" % key)
        setattr(self, key, value)

def toint(val):
    if not val or val == 'None':
        return 0
    return int(val)

def tofloat(val):
    if not val or val == 'None':
        return 0
    return float(val)

def todatetime(val):
    if not val or val == 'None':
        return datetime.datetime.min
    rt = re.findall(r'(\d{4})\W?(\d{2})\W?(\d{2})(?:[\WT]?(\d{2})\W?(\d{2})\W?(\d{2}))?', val)
    if not rt:
        rt = re.findall(r'(\d{2,4})\W(\d{1,2})\W(\d{1,2})(?:[\WT](\d{1,2})\W(\d{1,2})\W(\d{1,2}))?', val)
    if rt:
        t = [v if v else '00' for v in rt[0]]
        val = '{0}-{1}-{2} {3}:{4}:{5}'.format(t[0], t[1], t[2], t[3], t[4], t[5])
    else:
        return datetime.datetime.min
    return datetime.datetime.strptime(val, '%Y-%m-%d %H:%M:%S')

def toseconds(val):
    dt = todatetime(val)
    if dt is datetime.datetime.min:
        return False
    return dt.timestamp()

def to_human_time(val, fmt=None):
    if not fmt:
        fmt = '%Y-%m-%d %H:%M:%S'
    if not val:
        return ''
    return time.strftime(fmt, time.localtime(val))

def md5_encode(val):
    m = hashlib.md5(val.encode())
    return m.hexdigest()

def base64_encode(val):
    if isinstance(val, bytes):
        return base64.b64encode(val).decode()
    return base64.b64encode(val.encode()).decode()

def base64_decode(val):
    return base64.b64decode(val).decode()

def url_encode(val):
    return urllib.parse.quote(val)

def url_decode(val):
    return urllib.parse.unquote(val)

def random_string(len = 16):
    return ''.join(random.sample(_CHARACTORS, len))

def get_host_from_url(val):
    pos = 0
    if val.startswith('http'):
        pos = 8
    pos = val.find('/', pos)
    if pos >= 0:
        return val[:pos]
    return val

def parse_remote_ip(request):
    ip = request.headers.get('X-Real-Ip')
    if not ip:
        ip = request.remote_ip
    return ip

def format_url_with_params(url, params):
    slices = [str(k)+'='+urllib.parse.quote(str(v).encode()) for k, v in params.items()]
    sep = '&' if '?' in url else '?'
    return url + sep + '&'.join(slices)

def prepare_requests_certs_params(conf):
    if conf and 'cert' in conf and 'key' in conf:
        return (conf.get('cert'), conf.get('key'))
    return None

if __name__ == '__main__':
    # todatetime('2018/01/02T03:04:05:Z')
    # todatetime('2018-01-02 03:04:05')
    dt = todatetime('2018-01-02')
    print(dt)
    dt = todatetime('20180102')
    print(dt)
    dt = todatetime('2018-1-2T3:4:5')
    print(dt)
    ts = toseconds('2018-1-2T3:4:5')
    print(ts)
    ts = toseconds('1239-1-1')
    print(ts)
    ts = toseconds('123-1-a')
    print(ts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import time
from typing import Optional
from .supports import singleton
from .elasticsearchsaver import ELASTICSEARCH_SAVER

LOG = logging.getLogger('async.request')

@singleton
class ExceptionReporter(object):
    """Configures address that would reports execption content to
    """
    def __init__(self):
        super().__init__()
        self.db_connector = 'es-should-be-configured'
        self.db_category = 'exceptions'
        self.address = '127.0.0.1:9200'
        self.es_index = 'exceptions'
        self.es_doc_type = '_doc'

    def configure(self, category: str, conf: dict) -> bool:
        connector = conf.get('connector', '')
        host = conf.get('host', '127.0.0.1')
        port = int(conf.get('port', 0))
        configure_result = True
        if 'es' == connector or 'elasticsearch' == connector:
            self.db_connector = 'es'
            self.address = '%s:%d' % (host, port)
            # self.es_doc_type = category
            configure_result = ELASTICSEARCH_SAVER.configure(self.address)
            if configure_result:
                # mappings = {
                #     'index_patterns': [self.es_index+'-*'],
                #     'template': {
                #         'mappings': {
                #             'properties': {
                #                 'key': { 'type': 'text' },
                #                 'type': { 'type': 'text' },
                #                 'level': { 'type': 'text' },
                #                 'endpoint': { 'type': 'text' },
                #                 'method': { 'type': 'text' },
                #                 '@timestamp': {
                #                     'type': 'date',
                #                 },
                #                 'timestamp': {
                #                     'type': 'date',
                #                 }
                #             }
                #         },
                #     }
                # }
                mappings = self.get_elasticsearch_template(self.es_index)
                configure_result = ELASTICSEARCH_SAVER.ensure_template(index=self.es_index, mappings=mappings)
        elif 'restful' == connector or 'rest' == connector:
            self.db_connector = 'restful'
            self.address = host
        else:
            LOG.error('configure exception reporter failed while unknown connector:%s', connector)
            return False
        self.db_category = category
        return configure_result

    def report(self, key: str, typ: str, endpoint: str, method: str, inputs: str, outputs: str, content: str, level: str = 'ERROR', extra: Optional[dict] = None) -> bool:
        cur_date_endfix = '-' + time.strftime('%Y-%m-%d', time.localtime())
        data = {
            'key': key,
            'type': typ,
            'level': level,
            'endpoint': endpoint,
            'method': method,
            'input': inputs,
            'output': outputs,
            'content': content,
            'timestamp': time.time()*1000
        }
        if extra:
            for k, v in extra.items():
                if k not in data:
                    data[k] = v

        if 'es' == self.db_connector:
            data['@timestamp'] = data['timestamp']
            return ELASTICSEARCH_SAVER.save(index=self.es_index+cur_date_endfix, doc_type=self.es_doc_type, body=data)
        else:
            # not supported
            log_texts = ['%s:[%s]' % (k, k) for k, v in data.items()]
            LOG.error('reports of %s', ' '.join(log_texts))
            return False
        
    def get_elasticsearch_template(self, index: str) -> dict:
        mappings = {
            "order" : 0,
            "template": index + "-*",
            "index_patterns": [index + "-*"],
            "settings": {
                "index": {
                    "refresh_interval": "5s"
                }
            },
            "mappings": {
                "properties" : {
                    "@timestamp" : {
                        "type" : "date"
                    },
                    "@version" : {
                        "type" : "text"
                    },
                    'key': { 'type': 'text' },
                    'type': { 'type': 'text' },
                    'level': { 'type': 'text' },
                    'endpoint': { 'type': 'text' },
                    'method': { 'type': 'text' },
                    'input': { 'type': 'text' },
                    'output': { 'type': 'text' },
                    'timestamp': {
                        'type': 'date',
                    }
                }
            },
            "aliases" : { }
        }
        return mappings



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=C0111,C0103,R0205
"""
RabbitMQ With Tornado Connection
"""

import sys
import logging
import pika
from typing import Optional, Dict, List, Any
from pika.adapters.tornado_connection import TornadoConnection
import pika.channel
import pika.spec
import tornado.gen
import functools
import uuid
import socket
import json
import time
import copy
# import gevent
# import asyncio
from .supports import singleton, Constant
from .utilities import random_string
from .exceptionreporter import ExceptionReporter

logging.getLogger("apscheduler.executors.default").setLevel(logging.WARNING)
LOGGER = logging.getLogger('components.amqplib')

class _WorkerDelegateType(Constant):
    """
    """
    CONSUMER = 1
    PUBLISHER = 2

WORKER_TYPE = _WorkerDelegateType()

class _RabbitMQDefaults(Constant):
    """
    """
    SOCKET_TIMEOUT = 15
    HEARTBEAT_INTERVAL = 10

RABBIT_MQ_DEFAULTS = _RabbitMQDefaults()

class WorkerDelegate(object):
    """
    """
    worker_type = WORKER_TYPE.CONSUMER
    exchange = 'message'
    exchange_type = 'topic'
    exchange_durable = False
    queue = 'text'
    durable = False
    routing_key = 'example.text'
    callback = None
    auto_ack = True
    registered = False
    prefetch_count = 20

    def __init__(self, workerType, amqpProperties, callback):
        self.worker_type = workerType
        self.durable = amqpProperties.get('durable', True)
        self.exchange = amqpProperties.get('exchange')
        self.exchange_type = amqpProperties.get('exchange_type')
        self.exchange_durable = amqpProperties.get('exchange_durable', self.durable)
        self.queue = amqpProperties.get('queue')
        self.routing_key = amqpProperties.get('routing_key')
        self.auto_ack = amqpProperties.get('auto_ack')
        self.prefetch_count = 20
        self.worker_tag = '%s:%s:%s' % (self.exchange, self.routing_key, self.queue)
        self.consumer_tag = None
        self.registered = False

        if callable(callback):
            self.callback = callback
        else:
            self.callback = None

    @tornado.gen.coroutine
    def executor(self, worker, ch: pika.channel.Channel, basic_deliver: pika.spec.Basic.Deliver, properties: pika.BasicProperties, body: bytes):
        if callable(self.callback):
            worker._set_reply_needed_message(properties)
            try:
                if tornado.gen.is_coroutine_function(self.callback):
                    resp = yield self.callback(ch, basic_deliver, properties, body)
                else:
                    resp = self.callback(ch, basic_deliver, properties, body)
                if resp and properties.reply_to and properties.correlation_id:
                    if isinstance(resp, (list, dict)):
                        try:
                            resp = json.dumps(resp)
                        except:
                            LOGGER.error('json serialize %s failed with error:%s', str(resp), str(e))
                    if isinstance(resp, (str, bytes)):
                        if properties.correlation_id in worker._processing_replyies:
                            reply_properties = worker._processing_replyies.get(properties.correlation_id, properties)
                            reply_properties.timestamp = int(time.time())
                            worker.publish('', properties.reply_to, resp, reply_properties)
                return True
            except Exception as e:
                LOGGER.error('Execute consumer callback:%s failed with error:%s', self.callback.__name__, str(e))
                if properties.reply_to and properties.correlation_id and properties.correlation_id in worker._processing_replyies:
                    reply_properties = worker._processing_replyies.get(properties.correlation_id, properties)
                    reply_properties.timestamp = int(time.time())
                    resp = {'code': 500, 'message': str(e)}
                    worker.publish('', properties.reply_to, json.dumps(resp), reply_properties)
                inputs_text = body.decode('utf-8', errors='ignore') if isinstance(body, bytes) else str(body)
                ExceptionReporter().report(key='AMQP-'+str('consume'), typ='AMQP', 
                    endpoint='%s|%s|%s' % (str(basic_deliver.exchange), str(basic_deliver.routing_key), str(basic_deliver.consumer_tag)),
                    method='consume',
                    inputs=inputs_text,
                    outputs='',
                    content=str(e),
                    level='ERROR',
                    extra={
                        'worker_host': worker.worker_hostname,
                        'mq_properties': {
                            'correlation_id': properties.correlation_id if properties.correlation_id else '',
                            'reply_to': properties.reply_to if properties.reply_to else '',
                            'app_id': properties.app_id if properties.app_id else '',
                            'user_id': properties.user_id if properties.user_id else '',
                            'message_id': properties.message_id if properties.message_id else '',
                        }
                    }
                )
                return False
        return False

class PublishesMessage(object):
    """
    """
    exchange = 'message'
    routing_key = 'example.text'
    message: Optional[bytes] = None
    properties: Optional[pika.BasicProperties] = None
    callback: Optional[callable] = None

    def __init__(self, exchange: str, routing_key: str, body: bytes, properties: Optional[pika.BasicProperties] = None, callback: Optional[callable] = None):
        """Constructs a publishing message

        :param str exchange: The exchange name of destination rabbitmq
        :param str routing_key: The routing key of destination rabbitmq
        :param str|bytes body: The message content
        :param pika.BasicProperties properties: The rabbitmq message properties

        """
        self.exchange = exchange
        self.routing_key = routing_key
        self.message = body
        self.properties = properties
        self.callback = callback

class PublishesResponse(object):
    """RPC mq query response
    """

    def __init__(self, ch: pika.channel.Channel, basic_deliver: pika.spec.Basic.Deliver, properties: pika.BasicProperties, body: bytes):
        self.channel: pika.channel.Channel = ch
        self.delivery: pika.spec.Basic.Deliver = basic_deliver
        self.properties: pika.BasicProperties = properties
        self.body: bytes = body

class RabbitMQWorker(object):
    """
    """

    def __init__(self,
                 host='127.0.0.1',
                 port=5672,
                 virtual_host='/',
                 username='guest',
                 password='guest',
                 socket_timeout=RABBIT_MQ_DEFAULTS.SOCKET_TIMEOUT,
                 hb_interval=RABBIT_MQ_DEFAULTS.HEARTBEAT_INTERVAL,
                 publishing_interval=0.1):
        """Constructor

        :param str host: amqp host
        :param int port: amqp port
        :param str virtual_host: amqp virtual_host
        :param str username: amqp user name
        :param str password: amqp password
        :param str socket_timeout: amqp connecting socket timeout
        :param str hb_interval: amqp heartbeat interval

        """
        self._connection_info = 'amqp://%s@%s:%s/%s' % (username, host, str(port), virtual_host)
        LOGGER.debug("initializing amqp connection: %s", self._connection_info)

        super(RabbitMQWorker, self).__init__()

        credentials = pika.PlainCredentials(username, password)
        # parameter
        self._parameters = pika.ConnectionParameters(
            host=host, port=port, virtual_host=virtual_host,
            credentials=credentials, socket_timeout=socket_timeout, heartbeat=hb_interval)
        
        self.publishing_interval = publishing_interval
        self._connection: Optional[TornadoConnection] = None
        self._channel: Optional[pika.channel.Channel] = None
        self._closing = False
        self._custom_ioloop = None
        self._using_outside_ioloop = False
        self._main_thread_ioloop = None
        self._workers: Dict[str, WorkerDelegate] = {}
        self._workers_by_consumer_tag = {}
        self._publishes: List[PublishesMessage] = []
        self._publish_working = False
        self.tracker_queue_name = ''
        self.tracker_message_content = False
        self.worker_hostname = str(socket.gethostname())
        self._rpc_queue_name = 'rpc-%s-%s' % (self.worker_hostname, random_string(8))
        self._rpc_queue: Dict[str, PublishesMessage] = {}
        self._rpc_started: bool = False
        self._processing_replyies: Dict[str, pika.BasicProperties] = {}

    def set_custom_ioloop(self, custom_ioloop):
        """Specifies a custom ioloop, this method should be called before run

        :param None | asyncio.AbstractEventLoop |
            nbio_interface.AbstractIOServices custom_ioloop:
                Defaults to asyncio.get_event_loop().

        """
        self._custom_ioloop = custom_ioloop

    def set_main_thread_ioloop(self, main_thread_ioloop):
        """Specifies a ioloop, this method should be called before run

        :param None | asyncio.AbstractEventLoop |
            nbio_interface.AbstractIOServices main_thread_ioloop:
                Defaults to asyncio.get_event_loop().

        """
        self._main_thread_ioloop = main_thread_ioloop

    def set_using_outside_ioloop(self, using_outside_ioloop):
        """
        """
        self._using_outside_ioloop = using_outside_ioloop

    def connect(self) -> TornadoConnection :
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        :rtype: pika.adapters.tornado_connection.TornadoConnection

        """
        LOGGER.info('Connecting to %s', self._connection_info)
        self._closing = False
        return TornadoConnection(parameters=self._parameters,
                                 on_open_callback=self.on_connection_open,
                                 on_open_error_callback=self.on_connection_open_error,
                                 on_close_callback=self.on_connection_closed,
                                 custom_ioloop=self._custom_ioloop
                                )

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        LOGGER.info('Closing connection %s', self._connection_info)
        self._connection.close()

    def on_connection_closed(self, connection, reason):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.

        """
        self._channel = None
        for worker_cfg in self._workers.values():
            worker_cfg.registered = False
        self._publish_working = False
        if self._closing:
            self._connection.ioloop.stop()
        else:
            LOGGER.warning('Connection closed, reopening in 5 seconds: %s',
                           reason)
            self._connection.ioloop.call_later(5, self.reconnect)

    def on_connection_open(self, unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :param pika.SelectConnection _unused_connection: The connection

        """
        LOGGER.info('Connection opened')
        self.open_channel()

    def on_connection_open_error(self, connection, exception):
        """
        """
        LOGGER.error('Connecting to %s failed with error:%s, reconnecting', self._connection_info, str(exception))
        self._connection.ioloop.call_later(5, self.reconnect)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        if not self._closing:

            # Create a new connection
            self._connection = self.connect()

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        LOGGER.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed

        """
        LOGGER.warning('Channel %i was closed: %s', channel, reason)
        if self._connection.is_closed:
            LOGGER.warning('Channel %i closed while connection were close too.', channel)
            return
        self._connection.close()

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object

        """
        LOGGER.info('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.add_on_cancel_callback()

        self.setup_workers()

    def setup_exchange(self, worker_cfg):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        :param WorkerDelegate worker_cfg: Contains the name of the exchange to declare

        """
        LOGGER.info("Declaring exchange '%s'", worker_cfg.exchange)
        self._channel.exchange_declare(worker_cfg.exchange,
                                       worker_cfg.exchange_type,
                                       durable=worker_cfg.exchange_durable,
                                       callback=lambda unused_frame: self.on_exchange_declareok(unused_frame, worker_cfg)
                                      )

    def on_exchange_declareok(self, unused_frame, worker_cfg):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        :param str worker_cfg: WorkerDelegate config

        """
        LOGGER.info("Exchange '%s' declared", worker_cfg.exchange)
        self.setup_queue(worker_cfg)

    def setup_queue(self, worker_cfg: WorkerDelegate):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param WorkerDelegate worker_cfg: Contains the name of the queue to declare.

        """
        queue_name = worker_cfg.queue
        if worker_cfg.exchange_type == 'fanout':
            queue_name = ''
            LOGGER.info("Declaring queue '%s'", worker_cfg.queue)
            self._channel.queue_declare(queue_name, durable=False, auto_delete=True,
                                        callback= lambda method_frame : self.on_queue_declareok(method_frame, worker_cfg)
                                        )
        else:
            LOGGER.info("Declaring queue '%s'", worker_cfg.queue)
            self._channel.queue_declare(queue_name,
                                        durable=worker_cfg.durable,
                                        callback= lambda method_frame : self.on_queue_declareok(method_frame, worker_cfg)
                                        )

    def on_queue_declareok(self, method_frame, worker_cfg: WorkerDelegate):
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        :param pika.frame.Method method_frame: The Queue.DeclareOk frame
        :param WorkerDelegate worker_cfg: WorkerDelegate config

        """
        LOGGER.info('Declared queue:%s', method_frame.method.queue)
        if 'fanout' == worker_cfg.exchange_type:
            worker_cfg.queue = method_frame.method.queue
            LOGGER.info("Binding '%s' to '%s' with fanout type",
                        worker_cfg.exchange, worker_cfg.queue)
            self._channel.queue_bind(worker_cfg.queue, 
                                     worker_cfg.exchange, '', 
                                     callback= lambda unused_frame : self.on_bindok(unused_frame, worker_cfg)
                                    )
        else:
            LOGGER.info("Binding '%s' to '%s' with '%s'",
                        worker_cfg.exchange, worker_cfg.queue, str(worker_cfg.routing_key))
            binding_keys = worker_cfg.routing_key if (isinstance(worker_cfg.routing_key, list) or isinstance(worker_cfg.routing_key, tuple)) else [worker_cfg.routing_key]
            for binding_key in binding_keys:
                self._channel.queue_bind(worker_cfg.queue, 
                                         worker_cfg.exchange, 
                                         binding_key, 
                                         callback= lambda unused_frame : self.on_bindok(unused_frame, worker_cfg)
                                        )

    def on_bindok(self, unused_frame, worker_cfg: WorkerDelegate):
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method unused_frame: The Queue.BindOk response frame
        :param WorkerDelegate worker_cfg: WorkerDelegate config

        """
        LOGGER.info("Queue '%s' bound", worker_cfg.queue)
        if not worker_cfg.registered:
            # in case that binding a multiple routing key to the exchange
            worker_cfg.registered = True
            self.start_worker(worker_cfg)

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        LOGGER.info('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        LOGGER.info('Consumer was cancelled remotely, shutting down: %r',
                    method_frame)
        if self._channel:
            self._channel.close()

    def acknowledge_message(self, delivery_tag):
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.

        :param int delivery_tag: The delivery tag from the Basic.Deliver frame

        """
        # LOGGER.info('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def on_message(self, ch: pika.channel.Channel, basic_deliver: pika.spec.Basic.Deliver, properties: pika.BasicProperties, body: bytes):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel ch: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body
        :param WorkerDelegate worker_cfg: WorkerDelegate config

        """
        # LOGGER.info('Received message # %s from %s: %s', basic_deliver.delivery_tag, str(properties.app_id), body)
        worker_cfg = self._workers_by_consumer_tag.get(basic_deliver.consumer_tag)
        if worker_cfg and worker_cfg.callback:
            # self._connection.ioloop.call_soon(worker_cfg.callback, ch, basic_deliver, properties, body)
            # worker_cfg.callback(ch, basic_deliver, properties, body)
            self._connection.ioloop.add_callback(functools.partial(worker_cfg.executor, self, ch, basic_deliver, properties, body))
            if worker_cfg.auto_ack:
                self.acknowledge_message(basic_deliver.delivery_tag)

            if self.tracker_queue_name and properties.correlation_id:
                self._connection.ioloop.add_callback(functools.partial(self._publish_tracker), body, properties)
        else:
            self._channel.basic_nack(basic_deliver.delivery_tag, requeue=True)

    def on_cancel_consuming_ok(self, unused_frame):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method unused_frame: The Basic.CancelOk frame

        """
        LOGGER.info('RabbitMQ acknowledged the cancellation of the consumer')
        # TODO
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        LOGGER.info('Closing the channel')
        self._channel.close()

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        LOGGER.info('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        if self._channel:
            LOGGER.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            for worker_cfg in self._workers.values():
                if worker_cfg.consumer_tag:
                    self._channel.basic_cancel(worker_cfg.consumer_tag, self.on_cancel_consuming_ok)

    def start_worker(self, worker_cfg):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.

        :param WorkerDelegate worker_cfg: WorkerDelegate config

        """
        # self.add_on_cancel_callback()
        if worker_cfg.worker_type == WORKER_TYPE.CONSUMER:
            LOGGER.info("Issuing consumer for queue:'%s'", worker_cfg.queue)
            self._channel.basic_qos(prefetch_count=worker_cfg.prefetch_count)
            worker_cfg.consumer_tag = self._channel.basic_consume(worker_cfg.queue, 
                                                                  self.on_message,
                                                                  auto_ack = False)
            self._workers_by_consumer_tag[worker_cfg.consumer_tag] = worker_cfg
        
        elif not self._publish_working:
            self.start_publishing()

    def start_publishing(self):
        """This method will enable delivery confirmations and schedule the
        first message to be sent to RabbitMQ

        """
        LOGGER.info('Start publishing scheduler')
        # self.enable_delivery_confirmations()
        self.schedule_next_message()

    def enable_delivery_confirmations(self):
        """Send the Confirm.Select RPC method to RabbitMQ to enable delivery
        confirmations on the channel. The only way to turn this off is to close
        the channel and create a new one.

        When the message is confirmed from RabbitMQ, the
        on_delivery_confirmation method will be invoked passing in a Basic.Ack
        or Basic.Nack method from RabbitMQ that will indicate which messages it
        is confirming or rejecting.

        """
        LOGGER.info('Issuing Confirm.Select RPC command')
        self._channel.confirm_delivery(self.on_delivery_confirmation)

    def on_delivery_confirmation(self, method_frame):
        """Invoked by pika when RabbitMQ responds to a Basic.Publish RPC
        command, passing in either a Basic.Ack or Basic.Nack frame with
        the delivery tag of the message that was published. The delivery tag
        is an integer counter indicating the message number that was sent
        on the channel via Basic.Publish. Here we're just doing house keeping
        to keep track of stats and remove message numbers that we expect
        a delivery confirmation of from the list used to keep track of messages
        that are pending confirmation.

        :param pika.frame.Method method_frame: Basic.Ack or Basic.Nack frame

        """
        # confirmation_type = method_frame.method.NAME.split('.')[1].lower()
        # LOGGER.info('Received %s for delivery tag: %i', confirmation_type,
        #             method_frame.method.delivery_tag)
        # if confirmation_type == 'ack':
        #     self._acked += 1
        # elif confirmation_type == 'nack':
        #     self._nacked += 1
        # self._deliveries.remove(method_frame.method.delivery_tag)
        # LOGGER.info(
        #     'Published %i messages, %i have yet to be confirmed, '
        #     '%i were acked and %i were nacked', self._message_number,
        #     len(self._deliveries), self._acked, self._nacked)

    def schedule_next_message(self):
        """If we are not closing our connection to RabbitMQ, schedule another
        message to be delivered in publishing_interval seconds.

        """
        # LOGGER.info('Scheduling next message for %0.1f seconds', self.publishing_interval)
        if self._connection and self._publishes:
            self._publish_working = True
            self._connection.ioloop.add_callback(functools.partial(self._do_publish_message))
        else:
            self._publish_working = False

    @tornado.gen.coroutine
    def _do_publish_message(self):
        """If the class is not stopping, publish a message to RabbitMQ,
        appending a list of deliveries with the message number that was sent.
        This list will be used to check for delivery confirmations in the
        on_delivery_confirmations method.

        Once the message has been sent, schedule another message to be sent.
        The main reason I put scheduling in was just so you can get a good idea
        of how the process is flowing by slowing down and speeding up the
        delivery intervals by changing the publishing_interval constant in the
        class.

        """
        if self._channel is None or not self._channel.is_open:
            self._publish_working = False
            return

        for element in self._publishes:
            pub_properties = element.properties
            if callable(element.callback):
                if not element.properties:
                    correlation_id = str(uuid.uuid4())
                    element.properties = pika.BasicProperties(correlation_id=correlation_id,
                                                              reply_to=self._rpc_queue_name,
                                                              message_id=correlation_id,
                                                              headers={'X-Forward-For': self._rpc_queue_name}
                    )
                if not element.properties.correlation_id:
                    element.properties.correlation_id = str(uuid.uuid4())
                    if element.properties.reply_to:
                        if not element.properties.headers:
                            element.properties.headers = {'X-Forward-For': element.properties.reply_to}
                        elif 'X-Forward-For' not in element.properties.headers:
                            element.properties.headers['X-Forward-For'] = element.properties.reply_to
                element.properties.timestamp = int(time.time())
                pub_properties = copy.deepcopy(element.properties)
                if False == self._rpc_started:
                    self._ensure_rpc_response_queue()
                    self._rpc_started = True
                self._rpc_queue[pub_properties.correlation_id] = element
                pub_properties.reply_to = self._rpc_queue_name

            self._channel.basic_publish(element.exchange, element.routing_key,
                                        element.message,
                                        pub_properties)
            if self.tracker_queue_name and element.properties and element.properties.correlation_id:
                self._connection.ioloop.add_callback(functools.partial(self._publish_tracker), element.message, element.properties)
        self._publishes = []

        # hdrs = {u'مفتاح': u' قيمة', u'键': u'值', u'キー': u'値'}
        # properties = pika.BasicProperties(
        #     app_id='example-publisher',
        #     content_type='application/json',
        #     headers=hdrs)

        # message = u'مفتاح قيمة 键 值 キー 値'
        # self._channel.basic_publish(self.exchange, self.routing_key,
        #                             json.dumps(message, ensure_ascii=False),
        #                             properties)
        # self._message_number += 1
        # self._deliveries.append(self._message_number)
        # LOGGER.info('Published message # %i', self._message_number)

        if self._publish_working:
            self.schedule_next_message()

    def _set_reply_needed_message(self, properties: pika.BasicProperties):
        if properties.correlation_id and properties.reply_to:
            self._processing_replyies[properties.correlation_id] = properties
    
    def _answer_reply_needed_message(self, properties: pika.BasicProperties):
        if properties.correlation_id:
            if properties.correlation_id in self._processing_replyies:
                del self._processing_replyies[properties.correlation_id]

    def _is_reply_needed_message_answered(self, correlation_id: str) -> bool:
        if correlation_id:
            return correlation_id in self._processing_replyies
        return False

    def _ensure_rpc_response_queue(self):
        self._channel.queue_declare(self._rpc_queue_name, 
                                    durable=False,
                                    auto_delete=True,
                                    callback=lambda method: self._channel.basic_consume(self._rpc_queue_name, self._rpc_response, auto_ack=True)
                                    )

    @tornado.gen.coroutine
    def _rpc_response(self, ch: pika.channel.Channel, basic_deliver: pika.spec.Basic.Deliver, properties: pika.BasicProperties, body: bytes):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel ch: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body

        """
        if properties.correlation_id not in self._rpc_queue:
            return
        element = self._rpc_queue[properties.correlation_id]
        if callable(element.callback):
            element.callback(ch, basic_deliver, properties, body)
        del self._rpc_queue[properties.correlation_id]
        LOGGER.info('got rpc response %s', str(body))

    def run(self):
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.

        """
        try:
            self._connection = self.connect()
            if not self._using_outside_ioloop:
                self._connection.ioloop.start()
        except KeyboardInterrupt:
            self.stop()
            if (self._connection is not None and
                    not self._connection.is_closed):
                # Finish closing
                if not self._using_outside_ioloop:
                    self._connection.ioloop.start()

    def stop(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        LOGGER.info('Stopping')
        self._closing = True
        self.stop_consuming()
        self._connection.ioloop.stop()
        LOGGER.info('Stopped')

    def setup_workers(self):
        for worker_cfg in self._workers.values():
            if not worker_cfg.registered:
                self.setup_exchange(worker_cfg)
        
        if self._rpc_started:
            self._ensure_rpc_response_queue()

    def publish(self, exchange: str, routing_key: str, message: bytes, properties: Optional[pika.BasicProperties] = None, callback: Optional[callable] = None):
        """publishing a message

        :param str exchange: The exchange name of destination rabbitmq
        :param str routing_key: The routing key of destination rabbitmq
        :param str|bytes body: The message content
        :param pika.BasicProperties properties: The rabbitmq message properties

        """
        element = PublishesMessage(exchange,
                                   routing_key,
                                   message,
                                   properties,
                                   callback)
        self._publishes.append(element)

        if properties:
            # check if the publish message belongs to rpc reply
            if properties.correlation_id and routing_key == properties.reply_to:
                self._answer_reply_needed_message(properties)
        if not self._publish_working:
            self.start_publishing()

    def consume(self, exchange, exchange_type, binding_key, queue, durable, callback, auto_ack = True, exchange_durable = None):
        """Consumes a message

        :param str exchange: The exchange name of destination rabbitmq
        :param str exchange_type: The exchange type of destination rabbitmq
        :param str binding_key: The routing key of destination rabbitmq
        :param str queue: The destination queue
        :param bool durable: If the queue is durable
        :param callable callback: callback(pika.frame.Method) for method
          Basic.ConsumeOk.
        :param bool auto_ack: Enable manually ack when set to False
        :param bool exchange_durable: if the exchange is durable

        """
        key = '%s:%s:%s' % (exchange, binding_key, queue)
        if key in self._workers:
            LOGGER.error('Consumes a queue by exchange:%s binding_key:%s queue:%s while the consumer already registered.', exchange, binding_key, queue)
            return
        amqp_properties = {
            'exchange': exchange,
            'exchange_type': exchange_type,
            'exchange_durable': (durable if exchange_durable is None else exchange_durable),
            'routing_key': binding_key,
            'queue': queue,
            'durable': durable,
            'auto_ack': auto_ack,
        }
        worker_cfg = WorkerDelegate(WORKER_TYPE.CONSUMER, 
                                    amqp_properties,
                                    callback)
        self._workers[key] = worker_cfg
        if self._channel:
            self.setup_workers()

    def initialize_publisher(self, exchange, exchange_type, binding_key, queue, durable, exchange_durable = None):
        """Initialize a publishing queue and exchange

        :param str exchange: The exchange name of destination rabbitmq
        :param str exchange_type: The exchange type of destination rabbitmq
        :param str binding_key: The routing key of destination rabbitmq
        :param str queue: The destination queue
        :param bool durable: If the queue is durable
        :param bool exchange_durable: if the exchange is durable

        """
        key = '%s:%s:%s' % (exchange, binding_key, queue)
        if key in self._workers:
            LOGGER.error('Initialize a publisher by exchange:%s binding_key:%s queue:%s while the consumer already registered.', exchange, binding_key, queue)
            return
        amqp_properties = {
            'exchange': exchange,
            'exchange_type': exchange_type,
            'exchange_durable': (durable if exchange_durable is None else exchange_durable),
            'routing_key': binding_key,
            'queue': queue,
            'durable': durable,
            'auto_ack': False,
        }
        worker_cfg = WorkerDelegate(WORKER_TYPE.PUBLISHER, 
                                    amqp_properties,
                                    None)
        self._workers[key] = worker_cfg
        if self._channel:
            self.setup_workers()

    @tornado.gen.coroutine
    def _publish_tracker(self, message: bytes, properties: pika.BasicProperties):
        """If the class is not stopping, publish a message to RabbitMQ,
        appending a list of deliveries with the message number that was sent.
        This list will be used to check for delivery confirmations in the
        on_delivery_confirmations method.

        Once the message has been sent, schedule another message to be sent.
        The main reason I put scheduling in was just so you can get a good idea
        of how the process is flowing by slowing down and speeding up the
        delivery intervals by changing the publishing_interval constant in the
        class.

        """
        if not self.tracker_queue_name or self._channel is None or not self._channel.is_open:
            return

        self._channel.basic_publish('', self.tracker_queue_name,
                                    message if self.tracker_message_content else bytes('', 'utf-8'),
                                    properties)

@singleton
class RabbitMQFactory(object):

    def __init__(self):
        """Constructs a RabbitMQ operation factory

        """
        self._instances: Dict[str, WorkerDelegate] = {}
        self._default_virtual_host = None
        self._custom_ioloop = None
        self._tracker_queue_name: str = ''
        self._tracker_message_content: bool = False

    def initialize(self, conn_info: Dict[str, Any]):
        """Initializing a RabbitMQ operation factory

        :param dict conn_info: Connection configurations

        """
        virtual_host = conn_info.get('virtual_host')
        if virtual_host in self._instances:
            return self._instances[virtual_host]
        socket_timeout = conn_info['sock_timeout'] if 'sock_timeout' in conn_info else RABBIT_MQ_DEFAULTS.SOCKET_TIMEOUT
        heartbeat_interval = conn_info['heartbeat'] if 'heartbeat' in conn_info else RABBIT_MQ_DEFAULTS.HEARTBEAT_INTERVAL
        amqp_instance = RabbitMQWorker(host=conn_info.get('host'),
                                        port=conn_info.get('port'),
                                        username=conn_info.get('username'),
                                        password=conn_info.get('password'),
                                        virtual_host=virtual_host,
                                        socket_timeout=socket_timeout,
                                        hb_interval=heartbeat_interval)
        amqp_instance.set_using_outside_ioloop(True)

        amqp_instance.tracker_queue_name = self._tracker_queue_name
        amqp_instance.tracker_message_content = self._tracker_message_content
        self._instances[virtual_host] = amqp_instance
        if not self._default_virtual_host:
            self._default_virtual_host = virtual_host

        return amqp_instance

    def set_custom_ioloop(self, custom_ioloop):
        """Specifies a custom ioloop, this method should be called before run

        :param None | asyncio.AbstractEventLoop |
            nbio_interface.AbstractIOServices custom_ioloop:
                Defaults to asyncio.get_event_loop().

        """
        self._custom_ioloop = custom_ioloop

    def set_tracker_queue_name(self, queue_name: str):
        """Setup global tracker queue name, this would affect all workers
        """
        self._tracker_queue_name = queue_name
        for inst in self._instances.values():
            inst.tracker_queue_name = self._tracker_queue_name

    def enable_tracker_message_content(self, enabled: bool):
        """Setup global tracker queue name, this would affect all workers
        """
        self._tracker_message_content = True if enabled else False
        for inst in self._instances.values():
            inst.tracker_message_content = self._tracker_message_content

    def consume(self, virtual_host: str, exchange: str, exchange_type: str, binding_key: str, queue: str, durable: bool, callback: callable, auto_ack: bool = True, exchange_durable: bool = None):
        """Consumes a message

        :param str virtual_host: The destination connection virtual host
        :param str exchange: The exchange name of destination rabbitmq
        :param str exchange_type: The exchange type of destination rabbitmq
        :param str binding_key: The routing key of destination rabbitmq
        :param str queue: The destination queue
        :param bool durable: If the queue is durable
        :param callable callback: callback(pika.frame.Method) for method
          Basic.ConsumeOk.
        :param bool auto_ack: Enable manually ack when set to False
        :param bool exchange_durable: if the exchange is durable

        """
        self._instances[virtual_host].consume(exchange, exchange_type, binding_key, queue, durable, callback, auto_ack, exchange_durable)

    def initialize_publisher(self, virtual_host: str, exchange: str, exchange_type: str, binding_key: str, queue: str, durable: bool, exchange_durable: bool = None):
        """Initialize a publishing queue and exchange

        :param str virtual_host: The destination connection virtual host
        :param str exchange: The exchange name of destination rabbitmq
        :param str exchange_type: The exchange type of destination rabbitmq
        :param str binding_key: The routing key of destination rabbitmq
        :param str queue: The destination queue
        :param bool durable: If the queue is durable
        :param bool exchange_durable: if the exchange is durable

        """
        self._instances[virtual_host].initialize_publisher(exchange, exchange_type, binding_key, queue, durable, exchange_durable)

    def publish(self, virtual_host: str, exchange: str, routing_key: str, message: bytes, properties: Optional[pika.BasicProperties] = None, callback: Optional[callable] = None):
        """publishing a message

        :param str virtual_host: The destination connection virtual host
        :param str exchange: The exchange name of destination rabbitmq
        :param str routing_key: The routing key of destination rabbitmq
        :param str|bytes body: The message content
        :param pika.BasicProperties properties: The rabbitmq message properties

        """
        self._instances[virtual_host].publish(exchange, routing_key, message, properties, callback)

    def query_mq(self, virtual_host: str, exchange: str, routing_key: str, message: bytes, properties: Optional[pika.BasicProperties] = None):
        """publishing a message and waiting response

        :param str virtual_host: The destination connection virtual host
        :param str exchange: The exchange name of destination rabbitmq
        :param str routing_key: The routing key of destination rabbitmq
        :param str|bytes body: The message content
        :param pika.BasicProperties properties: The rabbitmq message properties

        :return tornado.gen.Future of PublishesResponse

        """
        waiter = tornado.gen.Future()
        callback = lambda ch, delivery, properties, body: waiter.set_result(PublishesResponse(ch, delivery, properties, body))
        # todo: waiter timeout
        self._instances[virtual_host].publish(exchange, routing_key, message, properties, callback)
        return waiter

    def run(self):
        # asyncio.set_event_loop(asyncio.new_event_loop())
        for amqp_instance in self._instances.values():
            amqp_instance.run()

    def stop(self):
        for amqp_instance in self._instances.values():
            amqp_instance.stop()



# -*- coding: utf-8 -*-
# @Author	: starview.brotherbaby
# @Date		: 2020-09-08 22:58:27 
# @Last Modified by:   starview.brotherbaby
# @Last Modified time: 2020-09-08 22:58:27
# Thanks for your comments!

import io

import qrcode

from . import base64_encode


def qr_data_to_img_base64_data(qr_code, 
    version=1, error_correction=qrcode.constants.ERROR_CORRECT_L, box_size=10, border=4):
    qr = qrcode.QRCode(
        version=version,
        error_correction=error_correction,
        box_size=box_size,
        border=border
    )
    qr.add_data(qr_code)
    qr.make(fit=True)
    img = qr.make_image()
    img_buffer = io.BytesIO()
    img.save(img_buffer, format='PNG')
    qr_data = base64_encode(img_buffer.getvalue())
    qr_img_data = 'data:image/png;base64,' + qr_data
    return qr_img_data

def img_to_base64_data(img_path, format='PNG'):
    with open(img_path,"rb") as f:
        img_data = base64_encode(f.read())
        return 'data:image/png;base64,' + img_data
    return ''



#!/usr/bin/python
#-*- coding:utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lycium",
    version="0.0.13",
    author="kevinyjn",
    author_email="kevinyjn@foxmail.com",
    description="Common python package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitee.com/starview/lycium",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "pika",
        "redis",
        "gunicorn",
        "mongoengine",
        "elasticsearch",
        "sqlalchemy",
        "Cython",
        "pycrypto",
        "IPy",
        "requests",
        "tornado",
        "zeep",
        "motor",
        "blinker",
        "aredis",
        "sqlalchemy_aio",
        "pyopenssl",
        "rsa",
        "cx_Oracle",
        "pymssql",
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import logging
import re
import six
import sys
import traceback
import sqlalchemy
from sqlalchemy.orm import query, loading, attributes
from sqlalchemy_aio import ASYNCIO_STRATEGY
import sqlalchemy.databases as supported_engines
# from sqlalchemy.orm import sessionmaker
import motor.motor_tornado
import motor
import tornado.gen
from tornado.ioloop import IOLoop

# from tornado.ioloop import IOLoop
# target_module = __import__('tornado.concurrent')
# if not hasattr(target_module.concurrent, 'return_future'):
#     setattr(target_module.concurrent, 'return_future', tornado.gen.coroutine)
# from motorengine import connect as mongo_connect
import mongoengine
import pymongo

try:
    import pymysql
except ImportError:
    pass
try:
    import pymssql
except ImportError:
    pass

try:
    import cx_Oracle
except ImportError:
    cx_Oracle = {}

try:
    from psycopg2cffi import compat
    compat.register()
except ImportError:
    pass


from .supports import singleton
from .modelutils import model_columns, format_mongo_value, get_dbinstance_by_model, DEFAULT_SKIP_FIELDS
from .utilities import url_encode
from .exceptionreporter import ExceptionReporter

LOG = logging.getLogger('components.dbproxy')

@singleton
class DbProxy(object):
    """
    database agent component
    """
    def __init__(self):
        self.db_instances = {}
        self.db_engines = {}
        self.default_mongo_db_instance = None
        self.default_rdbms_db_instance = None
        self._motor_count_documents_name = 'count_documents'

        if hasattr(motor.MotorCollection, 'count'):
            self._motor_count_documents_name = 'count'

        self._cur_execution_dbinstances = []

    def setup_rdbms(self, rdbms_configs: dict) -> bool:
        for k, dbconf in rdbms_configs.items():
            self.setup_rdbms_connection(k, dbconf)

    def setup_mongodbs(self, mongodb_configs: dict) -> bool:
        for k, dbconf in mongodb_configs.items():
            self.setup_mongodb_elenment(k, dbconf)

    def setup_rdbms_connection(self, category: str, dbconf: dict) -> bool:
        engine = dbconf.get('connector', '')
        if engine not in supported_engines.__all__:
            return False
        if category.startswith('ignore'):
            return False

        create_engine_params = {
            "encoding": dbconf.get('encoding', 'utf8'),
            "pool_pre_ping": True, 
            "strategy": ASYNCIO_STRATEGY
        }
        if 'connect_args' in dbconf:
            create_engine_params["connect_args"] = dbconf['connect_args']
        conndsn = self.format_connection_string(dbconf)
        db_engine = sqlalchemy.create_engine(conndsn, **create_engine_params)

        db_conn = IOLoop.current().run_sync(db_engine.connect)
        self.db_instances[category] = db_conn
        if not self.default_rdbms_db_instance:
            self.default_rdbms_db_instance = db_conn

        # syncronized connection
        self.db_engines[category] = sqlalchemy.create_engine(conndsn, 
            encoding=dbconf.get('encoding', 'utf8'),
            pool_pre_ping=True)

        return True

    def setup_mongodb_elenment(self, category: str, dbconf: dict) -> bool:
        if category.startswith('ignore'):
            return False
        conndsn = self.format_connection_string(dbconf)
        mongo_cli = motor.motor_tornado.MotorClient(conndsn)
        mongo_db = mongo_cli[dbconf.get('db')]
        if not self.default_mongo_db_instance:
            self.default_mongo_db_instance = mongo_db
        self.db_instances[category] = mongo_db
        return True
    
    def format_connection_string(self, dbconf: dict) -> str:
        engine = dbconf.get('connector', 'mysql')
        user = dbconf.get('user', 'guest')
        pwd = dbconf.get('pwd', '')
        host = dbconf.get('host', 'localhost')
        port = int(dbconf.get('port', 3306))
        dbname = dbconf.get('db', 'guest')
        driverpart = dbconf.get('driver', '')
        if driverpart:
            driverpart = '+' + driverpart

        connection_str = ''
        if engine == 'oracle':
            if sys.platform == 'darwin':
                connection_str = "oracle://%s:%s@%s:%d/%s" % (url_encode(user), url_encode(pwd), host, port, dbname)
            else:
                ora_dsn = cx_Oracle.makedsn(host, port, service_name=dbname)
                connection_str = "oracle://%s:%s@%s" % (user, pwd, ora_dsn)
        else:
            connection_str = '%s%s://%s:%s@%s:%d/%s' % (engine, driverpart, url_encode(user), url_encode(pwd), host, port, dbname)

        return connection_str

    @tornado.gen.coroutine
    def exec_query(self, db_name, sql):
        """
        execute sql for orm dbs
        """
        ret_list = []
        if (not sql) or (db_name not in self.db_instances):
            return ret_list
        try:
            exec_result = yield self.db_instances[db_name].execute(sql)
            ret_list = yield exec_result.fetchall()
        except UnicodeDecodeError as e:
            LOG.error(traceback.format_exc())
            ExceptionReporter().report(key='SQL-'+str('query'), typ='SQL', 
                endpoint='%s|%s' % (str(db_name), str(sql)),
                method='QUERY',
                inputs=str(sql),
                outputs=str(ret_list),
                content=str(e),
                level='ERROR'
            )

        return ret_list

    @tornado.gen.coroutine
    def exec_update(self, db_name, sql):
        """
        execute sql for no back, update or insert
        """
        if (not sql) or (db_name not in self.db_instances):
            return False
        try:
            yield self.db_instances[db_name].execute(sql)
        except Exception as e:
            LOG.error(traceback.format_exc())
            ExceptionReporter().report(key='SQL-'+str('update'), typ='SQL', 
                endpoint='%s|%s' % (str(db_name), str(sql)),
                method='UPDATE',
                inputs=str(sql),
                outputs='',
                content=str(e),
                level='ERROR'
            )
            return False
        return True
    
    def call_procedure_oracle(self, db_name, proc_name, params, out_params=None):
        cur = self.db_engines[db_name].raw_connection().cursor()
        return_params = {}
        if out_params and isinstance(out_params, dict):
            for k, v in out_params.items():
                return_params[k] = cur.var(getattr(cx_Oracle, v))
            params.extend(return_params.values())
        if not isinstance(params, list):
            return False
        # 调用存储过程
        cur.callproc(proc_name, params)
        return {k:int(return_params[k].getvalue()) if out_params.get(k)=='NUMBER' and return_params[k].getvalue() is not None 
            else return_params[k].getvalue() for k in return_params}

    def get_mongo_dbinstance(self, model):
        return self.default_mongo_db_instance

    def get_model_dbinstance(self, model):
        modelName = str(model._sa_class_manager.class_.__name__)
        dbflag = get_dbinstance_by_model(modelName)
        if dbflag is None:
            return self.default_rdbms_db_instance
        if dbflag in self.db_instances:
            return self.db_instances[dbflag]
        else:
            LOG.error('query model:%s while could not determine the database instance', modelName)
        
        # return dbinstance
        return self.default_rdbms_db_instance

    # queries

    # common queries
    @tornado.gen.coroutine
    def query_list(self, model, filters, limit, offset, sort, direction, selections=None, joins=None):
        qry, dbinstance, columns, _ = self._format_rdbms_query(model, filters, sort, direction, joins=joins)
        
        total = yield self._execute_rdbms_query_count(dbinstance, qry)
        if not total:
            return [], total

        qry = qry.limit(limit).offset(offset)
        if selections:
            qry = qry.from_self(*selections)
        rows = yield self._execute_rdbms_result(dbinstance, qry, as_all=True)

        items = []
        if selections:
            for row in rows:
                item = {}
                for k in selections:
                    item[k.key] = row._row[row._keymap['anon_1_'+k.expression._key_label][2]]
                items.append(item)
        else:
            for row in rows:
                item = {}
                for k in columns:
                    if k in DEFAULT_SKIP_FIELDS:
                        continue
                    item[k] = row._row[row._keymap[getattr(model, k).expression._key_label][2]]
                items.append(item)

        return items, total

    def _format_rdbms_query(self, model, filters, sort, direction, joins=None):
        columns,pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        orderby = None
        if sort:
            if direction == 'desc':
                orderby = sqlalchemy.desc(getattr(model, sort))
            else:
                orderby = sqlalchemy.asc(getattr(model, sort))
        elif (dbinstance.dialect.name == 'mssql' or dbinstance.dialect.name == 'postgresql'):
            if pk:
                orderby = getattr(model, pk)
            else:
                orderby = getattr(model, columns[0])
        
        qry = query.Query(model)
        if joins:
            for join in joins:
                qry = qry.join(*join)
        qry = qry.filter(*filters)
        if isinstance(orderby, list):
            qry = qry.order_by(*orderby)
        elif orderby is not None:
            qry = qry.order_by(orderby)
        return qry, dbinstance, columns, pk

    @tornado.gen.coroutine
    def query_list_mongo(self, model, filters, limit, offset, sort, direction, selections=None, joins=None, as_dict=True):
        q = self._format_mongo_query(model, filters)
        if selections:
            selFields = {}
            for k in selections:
                selFields[k] = 1
            if selFields:
                q = q.fields(*selFields)
        if sort:
            if direction.lower() == 'desc':
                q = q.order_by('-'+sort)
            else:
                q = q.order_by(sort)
        if joins:
            # TODO
            pass

        collection = self._prepare_mongo_collection(model)
        total = yield getattr(collection, self._motor_count_documents_name)(q._query)

        q = q.skip(offset)

        cursor = collection.find(q._query)
        rows = yield cursor.to_list(length=limit)

        items = []
        if as_dict:
            if selections:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: format_mongo_value(v) for k, v in row.items() if k in selections}
                    items.append(item)
            else:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: format_mongo_value(v) for k, v in row.items() if k not in DEFAULT_SKIP_FIELDS}
                    items.append(item)
        else:
            conditions = {}
            prepare_items = {}
            for row in rows:
                item = model()
                for k, v in row.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            if attr not in conditions:
                                conditions[attr] = {'filters': {pk+'__in': []}, 'ref_doc_type': ref_doc_type}
                            if attr not in prepare_items:
                                prepare_items[attr] = {v: []}
                            elif v not in prepare_items[attr]:
                                prepare_items[attr][v] = []
                            conditions[attr]['filters'][pk+'__in'].append(v)
                            prepare_items[attr][v].append(item)
                    
                    setattr(item, attr, v)
                items.append(item)

            if conditions:
                for attr, ref_options in conditions.items():
                    ref_rows = yield self.query_all_mongo(ref_options['ref_doc_type'], ref_options['filters'], limit=limit, as_dict=as_dict)
                    for ref_row in ref_rows:
                        if ref_row.id in prepare_items[attr]:
                            for dst_item in prepare_items[attr][ref_row.id]:
                                setattr(dst_item, attr, ref_row)
        
        return items, total

    @tornado.gen.coroutine
    def query_all_mongo(self, model, filters, limit=100, sort=None, **kwargs):
        direction = kwargs.pop('direction', None)
        selections = kwargs.pop('selections', None)
        joins = kwargs.pop('joins', None)
        as_dict = kwargs.pop('as_dict', True)
        q = self._format_mongo_query(model, filters)
        if selections:
            selFields = {}
            for k in selections:
                selFields[k] = 1
            if selFields:
                q = q.fields(*selFields)
        if sort:
            if direction.lower() == 'desc':
                q = q.order_by('-'+sort)
            else:
                q = q.order_by(sort)
        if joins:
            # TODO
            pass

        collection = self._prepare_mongo_collection(model)

        cursor = collection.find(q._query)
        rows = yield cursor.to_list(length=limit)

        items = []
        if as_dict:
            if selections:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in row.items() if k in selections}
                    items.append(item)
            else:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in row.items() if k not in DEFAULT_SKIP_FIELDS}
                    items.append(item)
        else:
            conditions = {}
            prepare_items = {}
            for row in rows:
                item = model()
                for k, v in row.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            if attr not in conditions:
                                conditions[attr] = {'filters': {pk+'__in': []}, 'ref_doc_type': ref_doc_type}
                            if attr not in prepare_items:
                                prepare_items[attr] = {v: []}
                            elif v not in prepare_items[attr]:
                                prepare_items[attr][v] = []
                            conditions[attr]['filters'][pk+'__in'].append(v)
                            prepare_items[attr][v].append(item)
                    
                    setattr(item, attr, v)
                items.append(item)

            if conditions:
                for attr, ref_options in conditions.items():
                    ref_rows = yield self.query_all_mongo(ref_options['ref_doc_type'], ref_options['filters'], limit=limit, as_dict=as_dict)
                    for ref_row in ref_rows:
                        if ref_row.id in prepare_items[attr]:
                            for dst_item in prepare_items[attr][ref_row.id]:
                                setattr(dst_item, attr, ref_row)

        return items

    @tornado.gen.coroutine
    def find_one_mongo(self, model, *args, **kwargs):
        """
        """
        as_dict = kwargs.pop('as_dict', False)
        q = self._format_mongo_query(model, (args, kwargs))
        collection = self._prepare_mongo_collection(model)
        document = yield from collection.find_one(q._query)
        if not document:
            return None

        if as_dict:
            item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in document.items()}
            # for k1, k2 in model._reverse_db_field_map.items():
            #     item[k2] = format_mongo_value(document.get(k1))
            return item
        else:
            item = model()
            for k, v in document.items():
                attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                    field = model._fields.get(attr)
                    ref_doc_type = field.document_type
                    pk = ref_doc_type._reverse_db_field_map.get('_id')
                    if pk:
                        ref_filters = {pk: v, 'as_dict': as_dict}
                        v = yield self.find_one_mongo(ref_doc_type, **ref_filters)

                setattr(item, attr, v)

        return item

    @tornado.gen.coroutine
    def mongo_aggregate(self, model, pipeline=[], *args, **kwargs):
        """
        query mongo by aggregate.
        """
        as_dict = kwargs.pop('as_dict', True)
        collection = self._prepare_mongo_collection(model)
        cursor = collection.aggregate(pipeline)
        data_list = []
        if as_dict:
            while (yield cursor.fetch_next):
                item = cursor.next_object()
                data_list.append(item)
        else:
            while (yield cursor.fetch_next):
                item = cursor.next_object()
                obj = model()
                for k, v in item.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            ref_filters = {pk: v, 'as_dict': as_dict}
                            v = yield self.find_one_mongo(ref_doc_type, **ref_filters)

                    setattr(obj, attr, v)
                data_list.append(obj)

        return data_list

    @tornado.gen.coroutine
    def save_mongo(self, item, force_insert=False, validate=True, clean=True,
             write_concern=None, cascade=None, cascade_kwargs=None,
             _refs=None, save_condition=None, signal_kwargs=None, **kwargs):
        """Save the :class:`~mongoengine.Document` to the database. If the
        document already exists, it will be updated, otherwise it will be
        created.

        :param force_insert: only try to create a new document, don't allow
            updates of existing documents.
        :param validate: validates the document; set to ``False`` to skip.
        :param clean: call the document clean method, requires `validate` to be
            True.
        :param write_concern: Extra keyword arguments are passed down to
            :meth:`~pymongo.collection.Collection.save` OR
            :meth:`~pymongo.collection.Collection.insert`
            which will be used as options for the resultant
            ``getLastError`` command.  For example,
            ``save(..., write_concern={w: 2, fsync: True}, ...)`` will
            wait until at least two servers have recorded the write and
            will force an fsync on the primary server.
        :param cascade: Sets the flag for cascading saves.  You can set a
            default by setting "cascade" in the document __meta__
        :param cascade_kwargs: (optional) kwargs dictionary to be passed throw
            to cascading saves.  Implies ``cascade=True``.
        :param _refs: A list of processed references used in cascading saves
        :param save_condition: only perform save if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`OperationError` if the conditions are not satisfied
        :param signal_kwargs: (optional) kwargs dictionary to be passed to
            the signal calls.

        .. versionchanged:: 0.5
            In existing documents it only saves changed fields using
            set / unset.  Saves are cascaded and any
            :class:`~bson.dbref.DBRef` objects that have changes are
            saved as well.
        .. versionchanged:: 0.6
            Added cascading saves
        .. versionchanged:: 0.8
            Cascade saves are optional and default to False.  If you want
            fine grain control then you can turn off using document
            meta['cascade'] = True.  Also you can pass different kwargs to
            the cascade save using cascade_kwargs which overwrites the
            existing kwargs with custom values.
        .. versionchanged:: 0.8.5
            Optional save_condition that only overwrites existing documents
            if the condition is satisfied in the current db record.
        .. versionchanged:: 0.10
            :class:`OperationError` exception raised if save_condition fails.
        .. versionchanged:: 0.10.1
            :class: save_condition failure now raises a `SaveConditionError`
        .. versionchanged:: 0.10.7
            Add signal_kwargs argument
        """
        if item._meta.get('abstract'):
            raise mongoengine.InvalidDocumentError('Cannot save an abstract document.')

        signal_kwargs = signal_kwargs or {}
        mongoengine.signals.pre_save.send(item.__class__, document=item, **signal_kwargs)

        if validate:
            item.validate(clean=clean)

        if write_concern is None:
            write_concern = {'w': 1}

        doc = item.to_mongo()

        created = ('_id' not in doc or item._created or force_insert)

        mongoengine.signals.pre_save_post_validation.send(item.__class__, document=item,
                                              created=created, **signal_kwargs)
        # it might be refreshed by the pre_save_post_validation hook, e.g., for etag generation
        doc = item.to_mongo()

        if item._meta.get('auto_create_index', True):
            self.mongo_model_ensure_indexes(item)
            pass

        try:
            # Save a new document or update an existing one
            if created:
                object_id = yield self._mongo_save_create(item, doc, force_insert, write_concern)
            else:
                object_id, created = yield self._mongo_save_update(item, doc, save_condition,
                                                       write_concern)

            if cascade is None:
                cascade = (item._meta.get('cascade', False) or
                           cascade_kwargs is not None)

            if cascade:
                kwargs = {
                    'force_insert': force_insert,
                    'validate': validate,
                    'write_concern': write_concern,
                    'cascade': cascade
                }
                if cascade_kwargs:  # Allow granular control over cascades
                    kwargs.update(cascade_kwargs)
                kwargs['_refs'] = _refs
                yield self.mongo_cascade_save(item, **kwargs)

        except pymongo.errors.DuplicateKeyError as err:
            message = 'Tried to save duplicate unique keys (%s)'
            raise mongoengine.NotUniqueError(message % six.text_type(err))
        except pymongo.errors.OperationFailure as err:
            message = 'Could not save document (%s)'
            if re.match('^E1100[01] duplicate key', six.text_type(err)):
                # E11000 - duplicate key error index
                # E11001 - duplicate key on update
                message = 'Tried to save duplicate unique keys (%s)'
                raise mongoengine.NotUniqueError(message % six.text_type(err))
            raise mongoengine.OperationError(message % six.text_type(err))

        # Make sure we store the PK on this document now that it's saved
        id_field = item._meta['id_field']
        if created or id_field not in item._meta.get('shard_key', []):
            item[id_field] = item._fields[id_field].to_python(object_id)

        mongoengine.signals.post_save.send(item.__class__, document=item,
                               created=created, **signal_kwargs)

        item._clear_changed_fields()
        item._created = False

        return item

    @tornado.gen.coroutine
    def insert_mongo(self, model, values):
        """Supports Document, dict, or list(builk insert)

        :param mongoengine.Document|str model: The mongo schema or collection name.
        :param mongoengine.Document|dict|list values: the mongo document or dict or list value to be inserted.
        """
        collection = self._prepare_mongo_collection(model)
        result = False
        if isinstance(values, mongoengine.Document):
            result = yield self.save_mongo(values)
        elif isinstance(values, dict):
            result = yield collection.insert_one(values)
        else:
            inserts = []
            for v in values:
                if isinstance(v, dict):
                    inserts.append(v)
                elif isinstance(v, mongoengine.Document):
                    item = v.to_mongo()
                    inserts.append(item)
                else:
                    raise Exception("insert mongodb values should either be dict or Document type!")
            if inserts:
                result = yield collection.insert_many(inserts)
        return result

    @tornado.gen.coroutine
    def update_mongo(self, model, item, condition):
        """
        update mongo

        :param mongoengine.Document|str model: The mongo schema or collection name.
        :param dict item: update key and value.
        :param dict condition: update by this conditions.
        return int: updated count.
        """
        if not (condition and isinstance(item, dict) and ("id" or "_id") not in item):
            raise mongoengine.OperationError("Invalid input.")
        collection = self._prepare_mongo_collection(model)

        try:
            updated_ret = yield collection.update_many(condition, {"$set": item})
            updated_count = updated_ret.modified_count
        except Exception:
            updated_count = 0

        return updated_count

    @tornado.gen.coroutine
    def del_item_mongo(self, item, condition=None):
        """
        :param mongoengine.Document|str item: The mongo document or schema or collection name
        :param dict|None condition: the delete condition if condition specified, the field 
            should be provided when item is schame or collection name.
        """
        result = False
        collection = self._prepare_mongo_collection(item)
        if not condition:
            result = yield collection.delete_one({'_id': item.pk()})
        elif isinstance(condition, dict):
            result = yield collection.delete_many(condition)
        return result

    def _prepare_mongo_collection(self, model):
        db = self.get_mongo_dbinstance(model)
        if isinstance(model, str):
            collection_name = model
        else:
            collection_name = model._meta.get('collection')
            if not collection_name:
                collection_name = model._class_name
        collection = db[collection_name]
        return collection

    def _format_mongo_query(self, model, filters):
        q = mongoengine.queryset.QuerySet(model, None)
        qfilters = []
        kwfilters = {}
        if isinstance(filters, tuple):
            for f in filters:
                if isinstance(f, dict):
                    for k,v in f.items():
                        kwfilters[k] = v
                elif isinstance(f, list):
                    for v in f:
                        qfilters.append(v)
        elif isinstance(filters, dict):
            kwfilters = filters
        elif isinstance(filters, list):
            qfilters = filters

        if qfilters or kwfilters:
            q = q.filter(*qfilters, **kwfilters)
        elif filters:
            q = q.filter(filters)
        return q

    @tornado.gen.coroutine
    def mongo_model_ensure_indexes(self, model):
        """Checks the document meta data and ensures all the indexes exist.

        Global defaults can be set in the meta - see :doc:`guide/defining-documents`

        .. note:: You can disable automatic index creation by setting
                  `auto_create_index` to False in the documents meta data
        """
        background = model._meta.get('index_background', False)
        drop_dups = model._meta.get('index_drop_dups', False)
        index_opts = model._meta.get('index_opts') or {}
        index_cls = model._meta.get('index_cls', True)

        collection = self._prepare_mongo_collection(model)
        # 746: when connection is via mongos, the read preference is not necessarily an indication that
        # this code runs on a secondary
        if not collection.is_mongos and collection.read_preference > 1:
            return

        # determine if an index which we are creating includes
        # _cls as its first field; if so, we can avoid creating
        # an extra index on _cls, as mongodb will use the existing
        # index to service queries against _cls
        cls_indexed = False

        # Ensure document-defined indexes are created
        if model._meta['index_specs']:
            index_spec = model._meta['index_specs']
            for spec in index_spec:
                spec = spec.copy()
                fields = spec.pop('fields')
                cls_indexed = cls_indexed or mongoengine.document.includes_cls(fields)
                opts = index_opts.copy()
                opts.update(spec)

                # we shouldn't pass 'cls' to the collection.ensureIndex options
                # because of https://jira.mongodb.org/browse/SERVER-769
                if 'cls' in opts:
                    del opts['cls']

                if mongoengine.pymongo_support.IS_PYMONGO_GTE_37:
                    collection.create_index(fields, background=background, **opts)
                else:
                    collection.ensure_index(fields, background=background,
                                            drop_dups=drop_dups, **opts)

        # If _cls is being used (for polymorphism), it needs an index,
        # only if another index doesn't begin with _cls
        if index_cls and not cls_indexed and model._meta.get('allow_inheritance'):

            # we shouldn't pass 'cls' to the collection.ensureIndex options
            # because of https://jira.mongodb.org/browse/SERVER-769
            if 'cls' in index_opts:
                del index_opts['cls']

            if mongoengine.pymongo_support.IS_PYMONGO_GTE_37:
                collection.create_index('_cls', background=background,
                                        **index_opts)
            else:
                collection.ensure_index('_cls', background=background,
                                        **index_opts)

    @tornado.gen.coroutine
    def _mongo_save_create(self, item, doc, force_insert, write_concern):
        """Save a new document.

        Helper method, should only be used inside save().
        """
        collection = self._prepare_mongo_collection(item)
        if force_insert:
            insert_result = yield collection.insert_one(doc)
            return insert_result.inserted_id
        # insert_one will provoke UniqueError alongside save does not
        # therefore, it need to catch and call replace_one.
        if '_id' in doc:
            raw_object = yield collection.replace_one(
                {'_id': doc['_id']}, doc)
            if raw_object:
                return doc['_id']

        insert_result = yield collection.insert_one(doc)
        object_id = insert_result.inserted_id

        # In PyMongo 3.0, the save() call calls internally the _update() call
        # but they forget to return the _id value passed back, therefore getting it back here
        # Correct behaviour in 2.X and in 3.0.1+ versions
        if not object_id and pymongo.version_tuple == (3, 0):
            pk_as_mongo_obj = item._fields.get(item._meta['id_field']).to_mongo(item.pk)
            rc = yield self.find_one_mongo(item.__class__, pk=pk_as_mongo_obj)
            if rc:
                object_id = rc.pk
            # object_id = (
            #     item._qs.filter(pk=pk_as_mongo_obj).first() and
            #     item._qs.filter(pk=pk_as_mongo_obj).first().pk
            # )  # TODO doesn't this make 2 queries?

        return object_id

    @tornado.gen.coroutine
    def _mongo_save_update(self, item, doc, save_condition, write_concern):
        """Update an existing document.

        Helper method, should only be used inside save().
        """
        collection = self._prepare_mongo_collection(item)
        object_id = doc['_id']
        created = False

        select_dict = {}
        if save_condition is not None:
            select_dict = mongoengine.queryset.query(item.__class__, **save_condition)
            # select_dict = mongoengine.transform.query(item.__class__, **save_condition)

        select_dict['_id'] = object_id

        # Need to add shard key to query, or you get an error
        shard_key = item._meta.get('shard_key', tuple())
        for k in shard_key:
            path = item._lookup_field(k.split('.'))
            actual_key = [p.db_field for p in path]
            val = doc
            for ak in actual_key:
                val = val[ak]
            select_dict['.'.join(actual_key)] = val

        update_doc = item._get_update_doc()
        if update_doc:
            wc = write_concern.pop('w')
            upsert = save_condition is None
            last_error = yield collection.update_one(select_dict, update_doc,
                                                     upsert=upsert, **write_concern)
            if not upsert and last_error._UpdateResult__raw_result['n'] == 0:
                raise mongoengine.SaveConditionError('Race condition preventing'
                                         ' document update detected')
            if last_error is not None:
                updated_existing = last_error.raw_result.get('updatedExisting')
                if updated_existing is False:
                    created = True
                    # !!! This is bad, means we accidentally created a new,
                    # potentially corrupted document. See
                    # https://github.com/MongoEngine/mongoengine/issues/564

        return object_id, created

    @tornado.gen.coroutine
    def mongo_cascade_save(self, item, doc, **kwargs):
        """Recursively save any references and generic references on the
        document.
        """
        _refs = kwargs.get('_refs') or []

        ReferenceField = mongoengine.document._import_class('ReferenceField')
        GenericReferenceField = mongoengine.document._import_class('GenericReferenceField')

        for name, cls in list(item._fields.items()):
            if not isinstance(cls, (ReferenceField,
                                    GenericReferenceField)):
                continue

            ref = item._data.get(name)
            if not ref or isinstance(ref, mongoengine.DBRef):
                continue

            if not getattr(ref, '_changed_fields', True):
                continue

            ref_id = "%s,%s" % (ref.__class__.__name__, str(ref._data))
            if ref and ref_id not in _refs:
                _refs.append(ref_id)
                kwargs["_refs"] = _refs
                yield self.save_mongo(ref, **kwargs)
                ref._changed_fields = []

    @tornado.gen.coroutine
    def query_all(self, model, filters, sort=None, direction='asc', joins=None):
        qry, dbinstance, columns, _ = self._format_rdbms_query(model, filters, sort, direction, joins=joins)
        
        rows = yield self._execute_rdbms_result(dbinstance, qry, as_all=True)
        items = []
        for row in rows:
            item = {}
            for k in columns:
                if k in DEFAULT_SKIP_FIELDS:
                    continue
                item[k] = row._row[row._keymap[getattr(model, k).expression._key_label][2]]
            items.append(item)

        return items

    @tornado.gen.coroutine
    def _execute_rdbms_result(self, dbinstance, qry, as_all=True):
        querycontext = qry._compile_context()
        querycontext.statement.use_labels = True
        result = yield dbinstance.execute(querycontext.statement, qry._params)
        if as_all:
            ret = yield result.fetchall()
        else:
            ret = yield result.fetchone()
        return ret

    @tornado.gen.coroutine
    def _execute_rdbms_query_count(self, dbinstance, qry):
        col = sqlalchemy.sql.func.count(sqlalchemy.sql.literal_column("*"))
        qrycount = qry.from_self(col)
        querycontext = qrycount._compile_context()
        querycontext.statement.use_labels = True
        result = yield dbinstance.execute(querycontext.statement, qrycount._params)
        ret = yield result.fetchone()
        if ret:
            return ret[0]
        return 0

    @tornado.gen.coroutine
    def find_item(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        qry = query.Query(model).filter(*filters)
        row = yield self._execute_rdbms_result(dbinstance, qry, as_all=False)
        if not row:
            return None
        columns, _ = model_columns(model)
        item = model()
        for k in columns:
            setattr(item, k, row._row[row._keymap[getattr(model, k).expression._key_label][2]])
        return item

    @tornado.gen.coroutine
    def get_count(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        qry = query.Query(model).filter(*filters)
        count = yield self._execute_rdbms_query_count(dbinstance, qry)
        return count

    @tornado.gen.coroutine
    def update_values(self, model, filters, values):
        dbinstance = self.get_model_dbinstance(model)
        stmt = sqlalchemy.update(model).where(*filters).values(**values)
        result = yield dbinstance.execute(stmt)
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    @tornado.gen.coroutine
    def update_item(self, item):
        model = item.__class__
        columns, pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        values, defaults = self.get_rdbms_instance_update_values(item, model, columns, pk)
        if not values:
            return False
        stmt = sqlalchemy.update(model).where(getattr(model, pk)==getattr(item, pk)).values(**values)
        result = yield dbinstance.execute(stmt)
        if result and result.rowcount:
            for k, v in defaults.items():
                setattr(item, k, v)
            
        return item

    @tornado.gen.coroutine
    def insert_item(self, item, auto_flush=False):
        model = item.__class__
        columns, pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        values, defaults = self.get_rdbms_instance_insert_values(item, model, columns, pk)
        if not values:
            return False
        stmt = sqlalchemy.insert(model).values(**values)
        result = yield dbinstance.execute(stmt)
        if result and result.rowcount:
            for k, v in defaults.items():
                setattr(item, k, v)
            if result.inserted_primary_key:
                setattr(item, pk, result.inserted_primary_key[0])
            
        return item

    @tornado.gen.coroutine
    def del_item(self, item):
        model = item.__class__
        dbinstance = self.get_model_dbinstance(model)
        _, pk = model_columns(model)
        stmt = sqlalchemy.delete(model).where(getattr(model, pk)==getattr(item, pk))
        result = yield dbinstance.execute(stmt)
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    @tornado.gen.coroutine
    def del_items(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        stmt = sqlalchemy.delete(model).where(*filters)
        result = yield dbinstance.execute(stmt)
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    def get_rdbms_instance_insert_values(self, item, model, columns, pk):
        values = {}
        defaults = {}
        for col in columns:
            tbl_col_name = getattr(model, col).expression.name
            if col in item._sa_instance_state.unmodified:
                v = self.get_rdbms_instance_default_value(getattr(model, col).expression, True)
                if v is None and col == pk:
                    continue
                values[tbl_col_name] = v
                defaults[tbl_col_name] = v
            else:
                values[tbl_col_name] = getattr(item, col)
        return values, defaults

    def get_rdbms_instance_update_values(self, item, model, columns, pk):
        values = {}
        defaults = {}
        for col in columns:
            tbl_col_name = getattr(model, col).expression.name
            values[tbl_col_name] = getattr(item, col)
            v = self.get_rdbms_instance_default_value(getattr(model, col).expression, False)
            if v is not None:
                values[tbl_col_name] = v
                defaults[tbl_col_name] = v
            
        return values, defaults

    def get_rdbms_instance_default_value(self, column, is_insert):
        column_default = column.default if is_insert else column.onupdate
        if column_default is None:
            return None
        else:
            return self._exec_rdbms_default(column, column_default, column.type)

    def _exec_rdbms_default(self, column, default, type_):
        if default.is_sequence:
            raise Exception("sequence default column value is not supported")
            return None # self.fire_sequence(default, type_)
        elif default.is_callable:
            return default.arg(None)
        elif default.is_clause_element:
            # TODO: expensive branching here should be
            raise Exception("clause default column value is not supported")
            # # pulled into _exec_scalar()
            # conn = self.connection
            # if not default._arg_is_typed:
            #     default_arg = expression.type_coerce(default.arg, type_)
            # else:
            #     default_arg = default.arg
            # c = expression.select([default_arg]).compile(bind=conn)
            # return conn._execute_compiled(c, (), {}).scalar()
            return None
        else:
            return default.arg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import urllib
import http
import traceback
import tornado.httpclient
import tornado.gen
import tornado.escape
from tornado.ioloop import IOLoop
from concurrent.futures import ThreadPoolExecutor
from tornado.concurrent import Future, run_on_executor
import asyncio
import zeep
from .tornadozeep import TornadoAsyncTransport
from .exceptionreporter import ExceptionReporter

tornado.httpclient.AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

logging.getLogger('tornado.curl_httpclient').setLevel(logging.WARN)
logging.getLogger('zeep.wsdl.wsdl').setLevel(logging.INFO)
logging.getLogger('zeep.xsd.schema').setLevel(logging.INFO)
logging.getLogger('zeep.transports').setLevel(logging.INFO)

LOG = logging.getLogger('async.request')

@tornado.gen.coroutine
def async_get(url, params=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('GET', url, params, body=None, json=None, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body.decode()

@tornado.gen.coroutine
def async_get_bytes(url, params=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('GET', url, params, body=None, json=None, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body

@tornado.gen.coroutine
def async_post(url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('POST', url, params, body=body, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body.decode()

@tornado.gen.coroutine
def async_post_bytes(url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('POST', url, params, body=body, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body

@tornado.gen.coroutine
def async_post_json(url, params=None, json=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('POST', url, params, body=None, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    if response.code == http.HTTPStatus.OK:
        data = tornado.escape.json_decode(response.body.decode())
        return True, data
    res = str(response.reason) + ' ' + response.body.decode()
    return False, res

@tornado.gen.coroutine
def async_http_request(method, url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    proxies = kwargs.pop('proxies', None)
    if params:
        qrs = [k + '=' + urllib.parse.quote(v) for k,v in params.items() ]
        sep = '&' if '?' in url else '?'
        url += sep + '&'.join(qrs)
    if json:
        body = tornado.escape.json_encode(json)
        if not headers:
            headers = {}
        headers['Content-Type'] = 'application/json'
    
    request = tornado.httpclient.HTTPRequest(url, method=method, body=body, headers=headers, validate_cert=verify_cert, **kwargs)
    client = tornado.httpclient.AsyncHTTPClient()
    prepare_request_proxies(request, proxies)
    
    response = yield client.fetch(request, raise_error=False)
    if response.code != http.HTTPStatus.OK:
        ExceptionReporter().report(key='HTTP-'+str(response.code), typ='HTTPQuery', 
            endpoint=url,
            method=method,
            inputs=body if body else tornado.escape.json_encode(params) if params else '',
            outputs=str(response.body.decode()),
            content=str(response.body.decode()),
            level='ERROR'
        )
    return response

def prepare_request_proxies(request, proxies):
    if proxies:
        if 'host' in proxies and 'port' in proxies:
            request.proxy_host = proxies.get('host')
            request.proxy_port = int(proxies.get('port'))
        else:
            schema = 'http'
            if request.url.startswith('https'):
                schema = 'https'
            if schema in proxies:
                part = proxies.get(schema)[len(schema)+3:]
                parts = part.split(':')
                request.proxy_host = parts[0]
                request.proxy_port = int(parts[1]) if len(parts) > 1 else (80 if schema == 'http' else 443)
        
class AsyncSoapExecutor():
    """
    """
    def __init__(self, ioloop = None):
        self.executor = ThreadPoolExecutor(max_workers=50)
        self.io_loop = ioloop or asyncio.new_event_loop()

    @run_on_executor
    def soap_request(self, future, endpoint, action, params, **kwargs) -> (str, str):
        proxies = kwargs.pop('proxies', None)
        transport = kwargs.pop('transport', None)
        uuid = kwargs.pop('uuid', None)
        err = ''
        context = ''
        try:
            web_client = zeep.Client(endpoint, transport=transport)
            soap_method = web_client.service[action]
            if isinstance(params, dict):
                context = soap_method(**params)
            elif isinstance(params, list):
                context = soap_method(*params)
            else:
                context = soap_method(params)
        except Exception as e:
            LOG.error(traceback.format_exc())
            err = str(e)
            ExceptionReporter().report(key='SOAP-'+str('failed'), typ='SOAPQuery', 
                endpoint=endpoint,
                method=str(action),
                inputs=str(params),
                outputs=str(context),
                content=err,
                level='ERROR'
            )
            # context = '{"code": -1, "message": %s}' % str(e)
        if future:
            future.set_result(context)
        # print('result', context)
        LOG.debug('%s %s %d response:\n%s', endpoint, action, uuid, context)
        return context, err

@tornado.gen.coroutine
def async_soap_request(endpoint, action, params, **kwargs) -> (str, str):
    # proxies = kwargs.pop('proxies', None)
    # transport = kwargs.pop('transport', None)
    # if transport is None:
    #     transport = TornadoAsyncTransport()
    # web_client = zeep.Client(endpoint, transport=transport)
    # soap_method = web_client.service[action]
    # context = soap_method(params)

    soap_executor = AsyncSoapExecutor()
    response, err = yield soap_executor.soap_request(None, endpoint, action, params, **kwargs)
    return response, err
    # future = Future()
    # # IOLoop.current().add_callback(_soap_request, future, endpoint, action, params, **kwargs)
    # return future
    # url = soap_client.location()
    # request = tornado.httpclient.HTTPRequest(url, method='POST', body=str(context.envelope), headers=context.client.headers(), **kwargs)
    # tornado_client = tornado.httpclient.AsyncHTTPClient()
    # prepare_request_proxies(request, proxies)
    # response = yield tornado_client.fetch(request)
    # print(response.body.decode())
    # # response_parser = SoapClient(web_client, action)
    # if response.code == http.HTTPStatus.OK:
    #     pass
    # else:
    #     ''
    # return response.code, response.body.decode()

def main():
    import tornado.web
    from tornado.ioloop import IOLoop

    routes = [
    ]

    class GeneralTornadoHandler(tornado.web.RequestHandler):
        """
        """

        def initialize(self, callback, methods):
            self.callbacks = {}
            if isinstance(methods, str):
                methods = [methods]
            for method in methods:
                method = method.upper()
                self.callbacks[method] = callback

        @tornado.gen.coroutine
        def get(self):
            yield self._do_callback(self.callbacks.get('GET'))

        @tornado.gen.coroutine
        def post(self):
            yield self._do_callback(self.callbacks.get('POST'))
        
        @tornado.gen.coroutine
        def _do_callback(self, cb):
            if callable(cb):
                response = yield cb(self, self.request)
                if isinstance(response, str):
                    self.write(response)
                    self.finish()
                else:
                    print("====== response", response)
                    pass
            else:
                print('not callable cb:', cb)

    def tornado_route(rule, **options):
        def decorator(f):
            # endpoint = options.pop('endpoint', None)
            # if not endpoint:
            #     endpoint = f.__name__

            routes.append((rule, GeneralTornadoHandler, dict(callback=f, methods=options.pop('methods', ['GET']))))
            return f
        return decorator

    app2 = tornado.web.Application(routes)
    app2.listen(8021)
    
    IOLoop.instance().start()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# decorator of sigleton
def singleton(clsname):
    instances = {}
    def getinstance(*args,**kwargs):
        if clsname not in instances:
            instances[clsname] = clsname(*args,**kwargs)
        return instances[clsname]
    return getinstance

class Constant(object):
    class ConstError(TypeError) : pass
    
    def __setattr__(self, key, value):
        if hasattr(self, key):
            raise self.ConstError("Can't rebind const (%s)" % key)
        setattr(self, key, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tornado.web
from .asynchttphandler import routes

class WebApplication(tornado.web.Application):
    """
    """
    
    def __init__(self, handlers=None, default_host=None, transforms=None, **settings):
        if not handlers:
            handlers = routes.routes
        return super().__init__(handlers=handlers, default_host=default_host, transforms=transforms, **settings)
